import json
import uuid
from copy import deepcopy
from typing import Union, List, Any
from warnings import warn

from pandas import DataFrame

from canvasxpress.config.collection import CXConfigs
from canvasxpress.config.type import CXConfig
from canvasxpress.data.base import CXData
from canvasxpress.data.convert import CXHtmlConvertable
from canvasxpress.data.keypair import CXDictData
from canvasxpress.data.matrix import CXDataframeData, merge_dataframes_into_xyz_object
from canvasxpress.data.text import CXTextData
from canvasxpress.js.collection import CXEvents
from canvasxpress.js.function import CXEvent
from canvasxpress.util.template import render_from_template

# from deprecated import deprecated

_DEFAULT_JS_URL: str = "https://www.canvasxpress.org/dist/canvasXpress.min.js"
_DEFAULT_VERSIONED_JS_URL = "https://cdnjs.cloudflare.com/ajax/libs/canvasXpress/@cx_version@/canvasXpress.min.js"

_DEFAULT_CSS_URL: str = "https://www.canvasxpress.org/dist/canvasXpress.css"
_DEFAULT_VERSIONED_CSS_URL = (
    "https://cdnjs.cloudflare.com/ajax/libs/canvasXpress/@cx_version@/canvasXpress.css"
)

_CX_JS_TEMPLATE = "var cX@cx_target_id@ = new CanvasXpress(@cx_json@); @cx_functions@"
"""
The template for declaring a CanvasXpress Javascript object using data
from the Python edition.
"""

_CX_LICENSE_TEMPLATE = "<script src='@cx_license@' type='text/javascript'></script>"
"""
The template for declaring a CanvasXpress license file so that charts do 
not create_element with unlicensed watermarks.
"""

_CX_REPR_TEMPLATE = """CanvasXpress(
    render_to="@render_to@",
    data=@data@,
    config=@config@,
    width=@width@,
    height=@height@,
    events=@events@,
    after_render=@after_render@,
    other_init_params=@other_init_params@
)
"""


class CanvasXpress(CXHtmlConvertable):
    """
    CanvasXpress acts as a proxy to the Javascript CanvasXpress object, and in
    general use remains similar to its Javascript counterpart.

    Assuming a flask function that returns a rendered page using the data from
    a CanvasXpress object:

    ```python
    @app.route('/pythonexample')
    def get_simple_chart() -> str:
        chart: CanvasXpress = CanvasXpress(
            render_to="example_chart",
            data=CXDictData(
                {
                    "y": {
                    "vars": ["Gene1"],
                    "smps": ["Smp1", "Smp2", "Smp3"],
                    "data": [[10, 35, 88]],
                }
            ),
            config=CXConfigs(
                CXGraphType(CXGraphTypeOptions.Bar)
            )
        )

        html_parts = chart.render_to_html_parts()

        return render_template(
            "bar.html",
            canvas_element=html_parts["cx_canvas"],
            bar_graph=html_parts["cx_js"]
        )
    ```
    """

    __target_id: Union[str, None] = None
    """
    __target_id is used by the JS renderTo param.  None indicates that the object is anonymous.
    """

    @property
    def anonymous(self) -> bool:
        """
        Indicates whether the object is anonymous.  If `True`, then `render_to()` will result in a one-time random
        ID string being returned for use in contexts such as rapidly changing React environments.  If `False`, then
        the ID string returned is the one set for the object by the developer.
        :returns: `bool` `True` if the object is anonymous; otherwise `False`.
        """
        return self.__target_id is None

    @property
    def render_to(self) -> Union[str, None]:
        """
        The ID of the CanvasXpress object's associated HTML components, such as
        the create_element canvas element.  Sets the `id` attribute of the `<canvas>`
        element.
        :returns: `str` The ID, if configured; `None` if anonymous.
        """
        return (
            str(uuid.uuid4()).replace("-", "") if self.anonymous else self.__target_id
        )

    @render_to.setter
    def render_to(self, value: Union[str, None]) -> None:
        """
        Sets the render_to of the CanvasXpress instance.  Sets the `id`
        attribute of the `<canvas>` element.
        :param value:
            `str` The ID to be associated.  Must be alphanumeric.  Non-alphanumeric characters will
            be removed, except for `_`, and if the remaining string is empty then a UUID4 will be
            substituted.  This is to preserve JS compatibility during rendering. `None` can also
            be provided to indicate that this object should be anonymous, such as for use
            in rapidly changing React interfaces.
        """
        if not isinstance(value, str) and value is not None:
            raise TypeError("value must be of type str or None")

        elif value is not None:
            candidate = ""
            for c in value:
                if c.isalnum() or c == "_":
                    candidate += c
            if candidate == "":
                candidate = None

        else:
            candidate = value

        self.__target_id = candidate

    __license_url: Union[str, None] = None
    """
    Location of the CanvasXpressLicense.js file for use in rendering.
    """

    @property
    def license_available(self) -> bool:
        """
        Indicates if a license is associated with the CanvasXpress object.
        :returns: `True` if a license file URL has been set.
        """
        return self.__license_url is not None

    @property
    def license_url(self) -> str:
        """
        Returns the location of the license file associated with the
        CanvasXpress object.
        :returns: `str` URL of the file or `None` if no file is associated.
        """
        return self.__license_url

    @license_url.setter
    def license_url(self, value: str) -> None:
        """
        Sets the location of the license file to be associated with the
        CanvasXpress object.
        :param value:
            `str` The path to the license file or `None` if a previously set URL
            is no longer valid.
        """
        if value is None:
            self.__license_url = None

        else:
            candidate = str(value)
            if "CanvasXpressLicense.js" not in candidate:
                raise ValueError(
                    "CanvasXpress license files must be named "
                    "'CanvasXpressLicense.js'"
                )

            else:
                self.__license_url = candidate

    __cdn_edition: Union[str, None] = None
    """
    The edition of CanvasXpress to use.  None indicates that the latest edition available shall be used.
    Used as a class variable.
    """

    @classmethod
    def cdn_edition(cls) -> Union[str, None]:
        """
        Indicates the version of CanvasXpress being used.

        :returns: `Union[str, None]`: The Javascript CDN version used or None if the latest.
        """
        return cls.__cdn_edition

    @classmethod
    def set_cdn_edition(cls, value) -> None:
        """
        Sets the version of the Javascript CDN to use when generating HTML or React components.  See thr CDN page
        for CanvasXpress to identify the versions available: https://cdnjs.com/libraries/canvasXpress.

        :param value:
            `Union[str, None]` The CDN version, such as `38.1`, or None if the latest version is preferred.
        """
        cls.__cdn_edition = None if value is None else str(value)

    __js_url: str = _DEFAULT_JS_URL
    """
    The preferred JS library URL.  By default the standard URL.
    """

    @classmethod
    def js_library_url(cls) -> Union[str, None]:
        """
        Indicates the version of CanvasXpress being used.  This can be either a version number, such as for cdnjs, or
        a fully qualified URL to the CanvasXpress library.

        :returns: `str`: The Javascript CDN version used.
        """
        candidate_url = cls.__js_url
        if cls.__cdn_edition is not None:
            candidate_url = _DEFAULT_VERSIONED_JS_URL.replace(
                "@cx_version@", cls.__cdn_edition
            )

        return candidate_url

    @classmethod
    def set_js_library_url(cls, value) -> None:
        """
        Sets the non-versioned URL to use for the CanvasXpress library.

        :param value:
            The HTTP(S) URL providing the JS library.
        """
        cls.__js_url = _DEFAULT_JS_URL if value is None else str(value)

    __css_url: str = _DEFAULT_CSS_URL
    """
    The preferred CSS library URL.  By default the standard URL.
    """

    @classmethod
    def css_library_url(cls) -> str:
        """
        Indicates the version of CanvasXpress being used.  This can be either a version number, such as for cdncss, or
        a fully qualified URL to the CanvasXpress library.

        :returns: `str`: The Javascript CDN version used.
        """
        candidate_url = cls.__css_url
        if cls.__cdn_edition is not None:
            candidate_url = _DEFAULT_VERSIONED_CSS_URL.replace(
                "@cx_version@", cls.__cdn_edition
            )

        return candidate_url

    @classmethod
    def set_css_library_url(cls, value) -> None:
        """
        Sets the non-versioned URL to use for the CanvasXpress library.

        :param value:
            The HTTP(S) URL providing the JS library.
        """
        cls.__css_url = _DEFAULT_CSS_URL if value is None else str(value)

    CHART_WIDTH_DEFAULT: int = 500
    """
    Default width of the chart when rendered, such as into HTML.
    """

    __chart_width: int = CHART_WIDTH_DEFAULT
    """
    Preferred width of the chart when rendered, such as into HTML.
    """

    @property
    def width(self) -> int:
        """
        Indicates the preferred <canvas> Web element width when rendered.  This
        property is used to facilitate integration with Web containers such
        as Jupyter notebooks. Added to the `<canvas>` element, and also
        influences create_element containers for contexts such as Jupyter Notebooks.
        :returns: `int` The width
        """
        return self.__chart_width

    @width.setter
    def width(self, value: int):
        """
        Sets the preferred Web element width when rendered. Added to the
        `<canvas>` element, and also influences create_element containers for contexts
        such as Jupyter Notebooks.
        :param value: `int`
            The pixel count.  Cannot be `None` or less than `1`.
        """
        if value is None:
            raise ValueError("element_width cannot be None")

        elif not isinstance(value, int):
            raise TypeError("element_width must be an int.")

        elif value < 1:
            raise ValueError("element_width cannot be less than 1 pixel")

        else:
            self.__chart_width = value

    CHART_HEIGHT_DEFAULT: int = 500
    """
    Default height of the chart in pixels when rendered, such as into HTML.
    """

    __chart_height: int = CHART_HEIGHT_DEFAULT
    """
    Preferred height of the chart in pixels when rendered, such as into HTML.
    """

    @property
    def height(self) -> int:
        """
        Indicates the preferred Web element height when rendered.  This
        property is used to facilitate integration with Web containers such
        as Jupyter notebooks.  Added to the `<canvas>` element, and also
        influences create_element containers for contexts such as Jupyter Notebooks.
        :returns: `int` The pixel count
        """
        return self.__chart_height

    @height.setter
    def height(self, value: int):
        """
        Sets the preferred Web element height when rendered.  Added to the
        `<canvas>` element, and also influences create_element containers for contexts
        such as Jupyter Notebooks.
        :param value: `int`
        """
        if value is None:
            raise ValueError("element_height cannot be None")

        elif not isinstance(value, int):
            raise TypeError("element_height must be an int.")

        elif value < 1:
            raise ValueError("element_height cannot be less than 1 pixel")

        else:
            self.__chart_height = value

    __data: CXData = None
    """
    _data is used by the JS data param, and it can be of numerous forms.  This
    class provides mapping mechanisms to translate Python objects into 
    structures usable by CanvasXpress.
    """

    @property
    def data(self) -> CXData:
        """
        Provides access to the CXData associated with this CanvasXpress chart.
        :returns: `CXData` The data to be associated with the chart.
        """
        return self.__data

    @data.setter
    def data(self, value: Union[CXData, dict, DataFrame, str, None]) -> None:
        """
        Sets the CXData associated with this CanvasXpress chart.
        :param value: `Union[CXData, dict, DataFrame, str, None]`
            An object translatable into a CXData type. If the object is an
            instance of CXData then it will be tracked by the CanvasXpress
            object; otherwise, a new CXData object will be created to manage
            the content.
        """
        if value is None or isinstance(value, bool):
            self.__data = CXDictData()

        elif isinstance(value, CXData):
            self.__data = value

        elif isinstance(value, dict):
            self.__data = CXDictData(value)

        elif isinstance(value, DataFrame):
            self.__data = CXDataframeData(value)

        elif isinstance(value, str):
            self.__data = CXTextData(value)

        else:
            raise TypeError(
                "data must be of type CXData, dict, DataFrame, str, bool, or None"
            )

    __sample_annotation: CXData = None
    """
    Provides convenience support when working with DataFrames similar to R.
    """

    @property
    def sample_annotation(self) -> CXData:
        """
        Provides access to the CXData associated with this CanvasXpress chart.
        :returns: `CXData` The data to be associated with the chart.
        """
        return self.__sample_annotation

    @sample_annotation.setter
    def sample_annotation(self, value: Union[CXDataframeData, DataFrame, None]) -> None:
        """
        Sets the CXData associated with this CanvasXpress chart.
        :param value: `Union[CXData, dict, DataFrame, str, None]`
            An object translatable into a CXData type. If the object is an
            instance of CXData then it will be tracked by the CanvasXpress
            object; otherwise, a new CXData object will be created to manage
            the content.
        """
        if value is not None and not isinstance(self.data, CXDataframeData):
            raise ValueError(
                "The data property must be a DataFrame before this property can be set."
            )

        elif value is None or isinstance(value, bool):
            self.__sample_annotation = None

        elif isinstance(value, DataFrame):
            self.__sample_annotation = CXDataframeData(value)

        elif isinstance(value, CXDataframeData):
            self.__sample_annotation = CXDataframeData(value)

        else:
            raise TypeError("data must be of type CXDataframeData,  DataFrame, or None")

    __variable_annotation: CXData = None
    """
    Provides convenience support when working with DataFrames similar to R.
    """

    @property
    def variable_annotation(self) -> CXData:
        """
        Provides access to the CXData associated with this CanvasXpress chart.
        :returns: `CXData` The data to be associated with the chart.
        """
        return self.__variable_annotation

    @variable_annotation.setter
    def variable_annotation(
        self, value: Union[CXDataframeData, DataFrame, None]
    ) -> None:
        """
        Sets the CXData associated with this CanvasXpress chart.
        :param value: `Union[CXData, dict, DataFrame, str, None]`
            An object translatable into a CXData type. If the object is an
            instance of CXData then it will be tracked by the CanvasXpress
            object; otherwise, a new CXData object will be created to manage
            the content.
        """
        if value is not None and not isinstance(self.data, CXDataframeData):
            raise ValueError(
                "The data property must be a DataFrame before this property can be set."
            )

        elif value is None or isinstance(value, bool):
            self.__variable_annotation = None

        elif isinstance(value, DataFrame):
            self.__variable_annotation = CXDataframeData(value)

        elif isinstance(value, CXDataframeData):
            self.__variable_annotation = CXDataframeData(value)

        else:
            raise TypeError("data must be of type CXDataframeData,  DataFrame, or None")

    __events: CXEvents = None
    """
    __events is used by the JS events param.
    """

    @property
    def events(self) -> CXEvents:
        """
        Provides access to the CXEvents associated with this CanvasXpress chart.
        :returns: `CXEvents` The events to be associated with the chart.
        """
        return self.__events

    @events.setter
    def events(self, events: Union[CXEvents, List[CXEvent], None]) -> None:
        """
        Sets the CXEvents associated with this CanvasXpress chart.
        :param events:
            `CXEvents, List[CXEvent], None` An object translatable into a
            CXEvents type.  If the object is an instance of CXEvents then it
            will be tracked by the CanvasXpress object; otherwise, a new
            CXEvents object will be created to manage the content.
        """
        if not events:
            self.__events = CXEvents()

        elif isinstance(events, CXEvents):
            self.__events = events

        elif isinstance(events, list):
            self.__events = CXEvents(*events)

        else:
            raise TypeError("value must be List[CXEvent] or CXEvents")

    __config: CXConfigs = None
    """
    __config is used to compose the config settings for the CanvasXpress object.
    """

    @property
    def config(self) -> CXConfigs:
        """
        Provides access to the CXConfigs associated with this CanvasXpress chart.
        :returns: `CXConfigs` The config to be associated with the chart.
        """
        return self.__config

    @config.setter
    def config(
        self, value: Union[List[CXConfig], List[tuple], List[list], dict, CXConfigs]
    ):
        """
        Sets the CXConfigs associated with this CanvasXpress chart.
        :param value: `Union[
                List[CXConfig],
                List[tuple],
                List[list],
                dict,
                CXConfigs
            ]`
            An object translatable into a CXConfigs type.  If the object is an
            instance of CXConfigs then it will be
            tracked by the CanvasXpress object; otherwise, a new CXConfigs
            object will be created to manage the content.
        """
        if value is None:
            self.__config = CXConfigs()

        elif isinstance(value, list):
            self.__config = CXConfigs(*value)

        elif isinstance(value, dict):
            self.__config = CXConfigs(value)

        elif isinstance(value, CXConfigs):
            self.__config = value

        else:
            raise TypeError(
                "value must be one of Union[List[CXConfig], List[tuple], "
                "dict, CXConfigs, CXConfigs]"
            )

    __after_render: CXConfigs = None
    """
    __after_render is used to compose the afterRender settings for the 
    CanvasXpress object.
    """

    @property
    def after_render(self) -> CXConfigs:
        """
        Provides access to the CXConfigs associated with this CanvasXpress
        chart's afterRender property.
        :returns: `CXConfigs`
            The after_render configuration to be associated with the chart.
        """
        return self.__after_render

    @after_render.setter
    def after_render(
        self, value: Union[List[CXConfig], List[tuple], List[list], dict, CXConfigs]
    ):
        """
        Sets the CXConfigs associated with this CanvasXpress chart's afterRender
        property.
        :param value: `Union[List[CXConfig], List[tuple], dict, CXConfigs,
            CXConfigs]`
            An object translatable into a CXConfigs
            type.  If the object is an instance of CXConfigs then it will be
            tracked by the CanvasXpress object; otherwise, a new CXConfigs
            object will be created to manage the content.
        """
        if value is None:
            self.__after_render = CXConfigs()

        elif isinstance(value, list):
            self.__after_render = CXConfigs(*value)

        elif isinstance(value, dict):
            self.__after_render = CXConfigs(value)

        elif isinstance(value, CXConfigs):
            self.__after_render = value

        else:
            raise TypeError(
                "value must be one of Union[List[CXConfig], List[tuple], "
                "dict, CXConfigs, CXConfigs]"
            )

    __canvas_attributes = CXConfigs()
    """
    __canvas_attributes modify the <canvas> element generated by this object.
    """

    @property
    def other_init_params(self) -> CXConfigs:
        """
        Provides access to additional parameters that will be used with the
        `CanvasXpress` for Javascript constructor, such as `afterRencderInit`.

        :returns: `CXConfigs`
            `CXConfig` objects to be provided as additional Javascript
            constructor parameter values.
        """
        return self.__canvas_attributes

    @other_init_params.setter
    def other_init_params(
        self, value: Union[List[CXConfig], List[tuple], List[list], dict, CXConfigs]
    ):
        """
        Set the additional parameters to be used with the `CanvasXpress` for
        Javascript constructor, such as `afterRencderInit`. The following
        parameters will be ignored and the properties for this class should
        be used instead due to Python tier functionality useful to the
        developer:

        * `renderTo`: Use `CanvasXpress.render_to`
        * `data`: Use `CanvasXpress.data`
        * `config`: Use `CanvasXpress.config`
        * `afterRender`: Use `CanvasXpress.after_render`
        * `width`: Use `CanvasXpress.width`
        * `height`: Use `CanvasXpress.height`

        :param value: `Union[List[CXConfig], List[tuple], List[list], dict,
            CXConfigs]`
            An object translatable into a CXConfigs type.  If the object is an
            instance of CXConfigs then it will be tracked by the CanvasXpress
            object; otherwise, a new CXConfigs object will be created to manage
            the content.
        """
        if value is None:
            self.__canvas_attributes = CXConfigs()

        elif isinstance(value, list):
            self.__canvas_attributes = CXConfigs(*value)

        elif isinstance(value, dict):
            self.__canvas_attributes = CXConfigs(value)

        elif isinstance(value, CXConfigs):
            self.__canvas_attributes = value

        else:
            raise TypeError(
                "value must be one of List[CXConfig], List[tuple], List[list], "
                "dict, CXConfigs"
            )

        # Clean override properties
        disallowed_attributes = [
            "id",
            "width",
            "height",
            "data",
            "config",
            "afterRender",
            "renderTo",
        ]
        for attribute in disallowed_attributes:
            self.other_init_params.remove(attribute)

    @classmethod
    def from_reproducible_json(
        cls,
        cx_json: str,
        include_factory: bool = False,
        include_system: bool = False,
    ) -> "CanvasXpress":
        """
        Initializes a new `CanvasXpress` object using a reproducable research
        JSON saved from a CanvasXpress chart rendered in a Web browser.

        :param cx_json: `str`
            A valid reproducable research JSON typical of those created by
            CanvasXpress when running a Web browser.

        :param include_factory: `bool`
            Default `False`.  If `False` remove the `factory` attribute.

        :param include_system: `bool`
            Default `False`.  If `False` remove the `system` attribute.

        :returns: `CanvasXpress`
            Returns a new `CanvasXpress` object will all properties filled using
            the information provided by the reproducable research JSON.
        """
        try:
            cx_json_dict = json.loads(cx_json)

            if not include_factory:
                if cx_json_dict.get("factory"):
                    del cx_json_dict["factory"]

            if not include_system:
                if cx_json_dict.get("system"):
                    del cx_json_dict["system"]

            cx_render_to = cx_json_dict.get("renderTo")
            cx_data = cx_json_dict.get("data")
            cx_config = cx_json_dict.get("config")
            cx_after_render = cx_json_dict.get("afterRender")
            cx_width = cx_json_dict.get("width", 500)
            cx_height = cx_json_dict.get("height", 500)
            cx_other_init_params = [
                (str(param), cx_json_dict[param])
                for param in cx_json_dict.keys()
                if param
                not in [
                    "id",
                    "width",
                    "height",
                    "data",
                    "config",
                    "afterRender",
                    "renderTo",
                ]
            ]

            # Capure and remove setDimensions width, height
            setDimensions_indexes = list()
            for instruction_index, instruction in enumerate(cx_after_render):
                if isinstance(instruction, list):
                    if len(instruction) >= 2:
                        if instruction[0] == "setDimensions":
                            dimensions = instruction[1]
                            if isinstance(dimensions, list):
                                if len(dimensions) >= 2:
                                    if isinstance(dimensions[0], int):
                                        cx_width = dimensions[0]
                                    if isinstance(dimensions[1], int):
                                        cx_height = dimensions[1]
                            setDimensions_indexes.append(instruction_index)

            for index in reversed(setDimensions_indexes):
                del cx_after_render[index]

            candidate = CanvasXpress(
                render_to=cx_render_to,
                data=cx_data,
                config=cx_config,
                after_render=cx_after_render,
                other_init_params=cx_other_init_params,
                width=int(cx_width),
                height=int(cx_height),
            )

            return candidate

        except Exception as e:
            raise ValueError(
                "cx_json must be a valid CanvasXpress reproducable" " research JSON"
            )

    def __init__(
        self,
        render_to: str = None,
        data: Union[CXData, dict, DataFrame, str, None] = None,
        sample_annotation: Union[CXDataframeData, DataFrame, None] = None,
        variable_annotation: Union[CXDataframeData, DataFrame, None] = None,
        events: Union[List[CXEvent], CXEvents] = None,
        config: Union[List[CXConfig], List[tuple], dict, CXConfigs] = None,
        after_render: Union[List[CXConfig], List[tuple], dict, CXConfigs] = None,
        other_init_params: Union[List[CXConfig], List[tuple], dict, CXConfigs] = None,
        width: int = CHART_WIDTH_DEFAULT,
        height: int = CHART_HEIGHT_DEFAULT,
        **kwargs: Any,
    ) -> None:
        """
        Initializes a new CanvasXpress object.  Default values are provided for
        all parameters if values are not specified; otherwise the arguments are
        treated as if an appropriate setter were used.
        :param render_to: See `render_to` property, except that on default
            initialization the object will be assigned an UUID4 value.
        :param data: See the `data` property
        :param events: See the `events` property
        :param config: See the `config` property
        :param after_render: See the `after_render` property
        :param other_init_params: See the 'other_init_params` property
        :param width: See the `width` property
        :param height: See the `height` property
        :param kwargs: `Any`
            Additional keyword arguments for CanvasXpress. Secondary keywords are
            added to the `config` property, overriding any existing items of the
            same name. Primary keywords `renderTo` or `afterRender` are mapped to
            the corresponding property (`render_to` or `after_render`).
        """

        super().__init__()

        config_updated = CXConfigs()

        if isinstance(config, list):
            config_updated = deepcopy(CXConfigs(*config))

        elif isinstance(config, dict):
            config_updated = deepcopy(CXConfigs(config))

        elif isinstance(config, CXConfigs):
            config_updated = deepcopy(config)

        elif config is not None:
            raise TypeError(
                "config must be one of Union[List[CXConfig], List[tuple], "
                "dict, CXConfigs]"
            )

        existing_config_labels = [item.label for item in config_updated.configs]

        for key, value in kwargs.items():
            if key == "renderTo":
                if render_to is not None:
                    warn("`render_to` argument has been overridden by `renderTo`")

                render_to = value

            elif key == "afterRender":
                if after_render is not None:
                    warn("`after_render` argument has been overridden by `afterRender`")

                after_render = value

            else:
                if key in existing_config_labels:
                    config_updated.remove(key)

                config_updated.set_param(label=key, value=value)

        self.render_to = render_to
        self.data = data
        self.sample_annotation = sample_annotation
        self.variable_annotation = variable_annotation
        self.events = events
        self.config = config_updated
        self.after_render = after_render
        self.other_init_params = other_init_params
        self.width = width
        self.height = height

    def provide_data_object(self) -> CXData:
        if isinstance(self.data, CXDataframeData):
            return CXDictData(
                merge_dataframes_into_xyz_object(
                    self.data,
                    self.sample_annotation,
                    self.variable_annotation,
                )
            )

        else:
            return self.data

    def prepare_html_element_parts(self) -> dict:
        """
        Converts the CanvasXpress object into CanvasXpress element components in
        anticipation of further use in renderable objects or conversion into HTML.

        If the associated `CXData` is a type of `CXProfiledData` and a profile
        has yet to be assigned then a profile can be assigned according to the
        `CXConfig` labelled `graphType`.  If a profile is assigned but is not
        `CXRawProfile` then the `graphType` can be reassessed, and if
        appropriate a new profile better aligned to the data can be provided.

        :returns: `dict` A map of values in anticipation of further conversion
            into html or a renderable.
        """
        cx_element_params = {
            "renderTo": self.render_to,
            "data": self.provide_data_object().render_to_dict(),
            "config": self.config.render_to_dict(),
            "afterRender": self.after_render.render_to_list(),
            "otherParams": self.other_init_params.render_to_dict(),
            "events": "js_events",
            "width": self.width,
            "height": self.height,
        }

        # Support unique data without JSON data structure
        if cx_element_params["data"].get("raw"):
            cx_element_params["data"] = cx_element_params["data"]["raw"]

        cx_element_params["events"] = self.events.render_to_js()

        return cx_element_params

    def render_to_html_parts(self, fix_missing_profile: bool = True) -> dict:
        """
        Converts the CanvasXpress object into HTML5 complant script.

        If the associated `CXData` is a type of `CXProfiledData` and a profile
        has yet to be assigned then a profile can be assigned according to the
        `CXConfig` labelled `graphType`.  If a profile is assigned but is not
        `CXRawProfile` then the `graphType` can be reassessed, and if
        appropriate a new profile better aligned to the data can be provided.

        :returns: `dict` A map of values appropriate for use in HTML, such as
            via a jinja template typical of flask apps.

        Assuming a jinja template such as this:
        ```html
        <html>
            <head>
                <meta charset="UTF-8">
                <title>Flask CanvasXpress Example</title>
            </head>
            <body>
                <!-- 1. DOM element where the visualization will be appear -->
                {{canvas_element|safe}}

                <!-- 2. Include the CanvasXpress library -->
                <link
                    href='https://www.canvasxpress.org/dist/canvasXpress.css'
                    rel='stylesheet' type='text/css'/>
                <script
                    src='https://www.canvasxpress.org/dist/canvasXpress.min.js'
                    type='text/javascript'>
                </script>

                <!-- 3. Include script to initialize object -->
                <script type="text/javascript">
                    onReady(function () {
                        {{bar_graph|safe}}
                    })
                </script>
            </body>
        </html>
        ```

        A flask function could create_element a page with a chart such as:
        ```python
        @app.route('/pythonexample')
        def get_simple_chart() -> str:
            chart: CanvasXpress = CanvasXpress(
                render_to="example_chart",
                data=CXDictData(
                    {
                        "y": {
                        "vars": ["Gene1"],
                        "smps": ["Smp1", "Smp2", "Smp3"],
                        "data": [[10, 35, 88]],
                    }
                ),
                config=CXConfigs(
                    CXGraphType(CXGraphTypeOptions.Bar)
                )
            )

            html_parts = chart.render_to_html_parts()

            return render_template(
                "bar.html",
                canvas_element=html_parts["cx_canvas"],
                bar_graph=html_parts["cx_js"]
            )
        ```

        :param fix_missing_profile: `bool`
            Defaults to `True`.  If `True` then CXData used for the chart will
            be provided with a data profile appropriate to the `graphType`
            (or CXStandardProfile if no graphType is provided).  If `False`
            then no profile will be applied to those data objects without
            profiles.

        :param match_profile_to_graphtype: `bool`
            Defaults to `True`.  If `True` then the `graphType` will be
            inspected and an appropriate data profile will be applied to
            the data object.  If a profile of an appropriate type is already
            associated then nothing is changed.  If a CXRawProfile is associated
            then no change is made regardless of the paranmeter value.
            Missing profiles are ignored unless fix_missing_profile is also
            `True`.  If `False` then no change to the data profile will be made
            if a profile is already associated with the data object.
        """
        #  Capture the ID once to avoid anonymous object calls producing different IDs.
        render_id = self.render_to

        primary_params = {
            "renderTo": render_id,
            "data": self.provide_data_object().render_to_dict(),
            "config": self.config.render_to_dict(),
            "events": "js_events",
        }
        secondary_params = self.other_init_params.render_to_dict()
        canvasxpress = {**primary_params, **secondary_params}
        after_render_functions = []
        for fx in self.after_render.render_to_list():
            params = [json.dumps(p) for p in fx[1]]
            after_render_functions.append(
                f"CanvasXpress.$('{render_id}').{fx[0]}({', '.join(params)})"
            )

        # Support unique data without JSON data structure
        if canvasxpress["data"].get("raw"):
            canvasxpress["data"] = str(canvasxpress["data"]["raw"])

        cx_js = render_from_template(
            _CX_JS_TEMPLATE,
            {
                "cx_target_id": render_id,
                "cx_json": json.dumps(canvasxpress),
                "cx_functions": "\n" + "; ".join(after_render_functions) + ";\n",
            },
        )

        cx_js = cx_js.replace(
            '"js_events"',
            self.events.render_to_js(),
        )

        canvas_configs = CXConfigs(
            {
                "id": render_id,
                "width": self.width,
                "height": self.height,
            }
        )
        cx_canvas = (
            "<canvas "
            + " ".join(
                [
                    f"{str(config.label)}={json.dumps(config.value)}"
                    for config in canvas_configs.configs
                ]
            )
            + "></canvas>"
        )

        cx_license = render_from_template(
            _CX_LICENSE_TEMPLATE, {"cx_license": self.license_url}
        )

        cx_web_data = dict()
        cx_web_data["cx_js"] = cx_js
        cx_web_data["cx_canvas"] = cx_canvas
        if self.license_available:
            cx_web_data["cx_license"] = cx_license

        return cx_web_data

    def __str__(self) -> str:
        """
        *str* function.  Converts the `CanvasXpress` object into a JSON
         representation.
        :returns" `str`
            JSON form of the `CanvasXpress` object.
        """
        data = str(type(self.data)).split(".")[-1][:-2] if self.data else "None"

        return (
            f"CanvasXpress ({hex(id(self))}):"
            f" render_to '{self.render_to}';"
            f" data <{data}>;"
            f" config {len(self.config.configs)} item(s);"
            f" after_render {len(self.after_render.configs)} item(s));"
            f" other_init_params"
            f" {len(self.other_init_params.configs)} item(s);"
            f" events {len(self.events.events)} function(s)."
        )

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the `CanvasXpress` object into a pickle
        string that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """

        normalized_data = self.data.render_to_dict(config=self.config)
        if normalized_data.get("raw"):
            normalized_data = normalized_data["raw"]
        str_data = json.dumps(normalized_data)
        str_data_parts = str_data.split("\n")
        str_data = "\n".join(["    " + line for line in str_data_parts])[4:]

        str_config = json.dumps(self.config.render_to_dict())
        str_config_parts = str_config.split("\n")
        str_config = "\n".join(["    " + line for line in str_config_parts])[4:]

        str_after_render = json.dumps(self.after_render.render_to_list())
        str_after_render_parts = str_after_render.split("\n")
        str_after_render = "\n".join(
            ["    " + line for line in str_after_render_parts]
        )[4:]

        str_other_init_params = json.dumps(
            self.other_init_params.render_to_dict(), indent=4
        )
        str_other_init_params_parts = str_other_init_params.split("\n")
        str_other_init_params = "\n".join(
            ["    " + line for line in str_other_init_params_parts]
        )[4:]

        repr_str = (
            _CX_REPR_TEMPLATE.replace("@render_to@", self.render_to)
            .replace("@data@", str_data)
            .replace("@config@", str_config)
            .replace("@width@", json.dumps(self.width))
            .replace("@height@", json.dumps(self.height))
            .replace("@events@", repr(self.events))
            .replace("@after_render@", str_after_render)
            .replace("@other_init_params@", str_other_init_params)
            .replace("true", "True")
            .replace("false", "False")
            .replace("null", "None")
        )

        return repr_str



import streamlit.components.v1 as components

from canvasxpress.canvas import CanvasXpress

_cx_iframe_padding = 50

_cx_fx_template = """
<script type="text/javascript">
    @code@
</script>
"""

_cx_html_template = """
<html>
    <head>
        <meta charset="UTF-8">
        <title>CanvasXpress</title>

        <!-- 1. Include the CanvasXpress library -->
        @canvasxpress_license@
        <link 
                href='@css_url@' 
                rel='stylesheet' 
                type='text/css'
                referrerpolicy='origin-when-cross-origin'
        />
        <script 
                src='@js_url@' 
                type='text/javascript'>
                referrerpolicy='origin-when-cross-origin'
        </script>
    </head>
    <body>
        <!-- 3. DOM element where the visualization will be displayed -->
        @canvases@
        <!-- 2. Include script to initialize object -->
        @js_functions@
     </body>
</html>
"""


def _get_chart_display_code(cx: CanvasXpress) -> tuple:
    """
    Get the html display code and iframe dimensions for rendering a CanvasXpress
    object in a Streamlit application.
    :param cx: `CanvasXpress`
        The `CanvasXpress` object to be rendered.
    :returns: `tuple` of html, iframe_width, iframe_height
    """
    html_parts = cx.render_to_html_parts()
    canvases = html_parts["cx_canvas"]
    js_functions = _cx_fx_template.replace("@code@", html_parts["cx_js"])

    cx_license = ""
    if html_parts.get("cx_license"):
        cx_license = html_parts["cx_license"]

    css_url = CanvasXpress.css_library_url()
    js_url = CanvasXpress.js_library_url()

    html = (
        _cx_html_template.replace("@canvases@", canvases)
        .replace("@canvasxpress_license@", cx_license)
        .replace("@js_functions@", js_functions)
        .replace("@css_url@", css_url)
        .replace("@js_url@", js_url)
    )

    iframe_width = cx.width + _cx_iframe_padding
    iframe_height = cx.height + _cx_iframe_padding

    return html, iframe_width, iframe_height


def plot(cx: CanvasXpress) -> None:
    """
    Renders the provided CanvasXpress object for display in a Streamlit application.
    :param cx: `CanvasXpress`
        The `CanvasXpress` object to be rendered.
    :returns: `None` or raises a `TypeError` exception if `cx` is not a CanvasXpress
        object.
    """
    if cx is None:
        return None

    elif not isinstance(cx, CanvasXpress):
        raise TypeError(f"Argument 'cx' is not a CanvasXpress object")

    html, iframe_width, iframe_height = _get_chart_display_code(cx)

    components.html(html, width=iframe_width, height=iframe_height)



"""
This module provides functionality for detecting the platform context that is active, such as RStudio.
"""
import os
from os import environ

CANVASXPRESS_TARGET_CONTEXT: str = "CANVASXPRESS_TARGET_CONTEXT"
"""
The ENV variable that can be set to indicate which context should be targetted regardless of what is detected.
Valid options are:

- shiny
- jupyter
- dash
- streamlit
- browser
"""

CONTEXT_BROWSER: str = "browser"
CONTEXT_DASH: str = "dash"
CONTEXT_JUPYTER: str = "jupyter"
CONTEXT_RSTUDIO: str = "rstudio"
CONTEXT_SHINY: str = "shiny"
CONTEXT_STREAMLIT: str = "streamlit"
CONTEXT_UNKNOWN: str = "UNKNOWN"

VALID_CONTEXTS: list = [
    CONTEXT_BROWSER,
    CONTEXT_DASH,
    CONTEXT_JUPYTER,
    CONTEXT_RSTUDIO,
    CONTEXT_SHINY,
    CONTEXT_STREAMLIT,
]


def is_rstudio_active() -> bool:
    """
    Indicates if RStudio is active at the time of function call.
    :returns" `bool` `True` if the RStudio IDE is running.
    """
    try:
        import shiny

        return bool(environ.get("RSTUDIO", False))

    except ModuleNotFoundError:
        return False


def is_shiny_available() -> bool:
    """
    Indicates if shiny is available at the time of function call.
    :returns" `bool` `True` if shiny is available.
    """
    try:
        import shiny

        return os.environ.get("SHINY_HOST") is not None

    except ModuleNotFoundError:
        return False


def is_dash_available() -> bool:
    """
    Indicates if dash is available at the time of function call.
    :returns" `bool` `True` if dash is available.
    """
    try:
        import dash

        return dash.get_app() is not None

    except ModuleNotFoundError:
        return False

    except Exception:
        return False


def is_ipython_available() -> bool:
    """
    Indicates if IPython is available at the time of function call.
    :returns" `bool` `True` if IPython is available.
    """
    try:
        from IPython import get_ipython

        return get_ipython().__class__.__name__ != "NoneType"

    except ModuleNotFoundError:
        return False

    except NameError:
        return False


def is_streamlit_available() -> bool:
    """
    Indicates if streamlit is available at the time of function call.
    :returns" `bool` `True` if streamlit is available.
    """
    try:
        import streamlit

        return True

    except ModuleNotFoundError:
        return False


def get_target_context() -> str:
    """
    Indicates the runtime context that should be targetted for illustrating charts.
    :returns" `str` One of the options for VALID_CONTEXTS.
    """
    target = environ.get(CANVASXPRESS_TARGET_CONTEXT)

    if target is not None:
        if target in VALID_CONTEXTS:
            return target

    else:
        if is_ipython_available():
            return CONTEXT_JUPYTER
        if is_dash_available():
            return CONTEXT_DASH
        if is_rstudio_active():
            return CONTEXT_RSTUDIO
        if is_shiny_available():
            return CONTEXT_SHINY
        if is_streamlit_available():
            return CONTEXT_STREAMLIT

        return CONTEXT_UNKNOWN



import json
from copy import copy, deepcopy
from functools import total_ordering

from canvasxpress.data.convert import CXJavascriptConvertable
from canvasxpress.util.template import render_from_template

_CX_EVENT_TEMPLATE = "function(o, e, t){@cx_js_script@}"
"""
_CX_EVENT_TEMPLATE provides the function format expected by CanvasXpress.
"""


@total_ordering
class CXEvent(CXJavascriptConvertable):
    """
    CXEvent is a `CXJavascriptConvertable` that represents Javascript source to
    be associated with a CanvasXpress object.

    CanvasXpress provides hook functions for various events that are called as
    those events occur for the `div` element containing the rendered chart.
    The format is:

    ```javascript
    "event-name": function(o, e, t) {
        event code
    }
    ```

    With the following as an example:

    ```javascript
    "mousemove": function(o, e, t) {
        t.showInfoSpan(e, '<pre>' + t.prettyJSON(o) + '</pre>');
    }
    ```

    CXEvent handles the function template, so the developer only needs to supply
    the event name and the source.  Given the above example, the following
    creates the equivalent CXEvent:

    ```python
    event = CXEvent(
        id="mousemove",
        script="t.showInfoSpan(e, '<pre>' + t.prettyJSON(o) + '</pre>');"
    )
    ```

    No validations is performed for `id` or `script`.

    Read the [CanvasXpress documentation](https://www.canvasxpress.org/docs.html#events)
    for additional information.  Also see `CXEvents`.
    """

    __id: str = ""
    """
    The triggering JS react, such as mousemove.
    """

    __script: str = ""
    """
    The script to be placed inside of the CanvasXpress react function, which
    takes the general form:
    
    function(o, e, t) {
        // script logic goes here
    };
    """

    @property
    def id(self) -> str:
        """
        Provides access to the react ID.
        :returns: The ID as a string.
        """
        return self.__id

    @id.setter
    def id(self, value: str) -> None:
        """
        Sets the react ID, which is a keyword recognized by CanvasXpress.
        :param value: `str`
            The ID, which must be a string compliant object.  Cannot be `None`.
        """
        if value == None:
            raise TypeError("value cannot be None")
        else:
            self.__id = str(value)

    @property
    def script(self) -> str:
        """
        Provides access to the react script.
        :returns: `str` The Javascript source.
        """
        return self.__script

    @script.setter
    def script(self, value: str) -> None:
        """
        Sets the react script, which is logic that goes inside of the react
        function.  Functions take the form:

        ```javascript
        function(o, e, t) {
            // script logic goes here
        };
        ```

        The script can be assumed to have access to all DOM elements as proper,
        and it will be provided the parameters o, e, and t.  Read the
        [CanvasXpress documentation](https://www.canvasxpress.org/docs.html#events)
        for additional information.

        :param value: `str`
            The ID, which must be a UTF-8 string compliant object.
        """
        if value == None:
            raise TypeError("value cannot be None")
        else:
            self.__script = str(value)

    def render_to_js(self) -> str:
        """
        Converts the object into HTML5 complant script.
        :returns: 'str'
        Given:
        ```python
        event1 = CXEvent("f1", "x = 0")
        function = event1.render_to_js()
        ```
        Then the value of `function` would be:
        ```text
        'f1': 'function(o, e, t){x = 0}
        ```
        """
        cx_js = self.script
        cx_js_func = render_from_template(
            _CX_EVENT_TEMPLATE,
            {
                "cx_js_script": cx_js,
            },
        )
        return cx_js_func

    def __init__(self, id: str = "", script: str = ""):
        """
        Initializes a new CXEvent object.
        :param id: `str`
            The ID of the react, such as mousemove.  Also see property `id`.
        :param script: `str`
            The script logic for the react.  Also see property `script`.
        """
        super().__init__()

        self.id = id
        self.script = script

    def __copy__(self):
        """
        *copy* constructor.  Returns a new `CXEvent` object.
        """
        return CXEvent(id=copy(self.id), script=copy(self.script))

    def __deepcopy__(self, memo):
        """
        *deepcopy* constructor.  Returns a new `CXEvent` object.
        """
        return CXEvent(id=deepcopy(self.id), script=deepcopy(self.script))

    def __lt__(self, other: "CXEvent") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXEvent` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXEvent` object then `False`
            <li> If `other` is a `CXEvent` object then True if id and string of
                `other` are less than that of `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXEvent):
            return False

        else:
            if self.id < other.id:
                return True
            elif self.id == other.id:
                return self.script < other.script
            else:
                return False

    def __eq__(self, other: "CXEvent") -> bool:
        """
        *equal* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXEvent` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXEvent` object then `False`
            <li> If `other` is a `CXEvent` object then True if id and string of
                `other` are equal to that of `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXEvent):
            return False

        else:
            return (self.id == other.id) and (self.script == other.script)

    def __str__(self) -> str:
        """
        *str* function.  Converts the object into a Javascript statement.
        """
        return f'"{self.id}": {self.render_to_js()}'

    @staticmethod
    def __clean_string(value: str) -> str:
        """
        Adjusts characters problematic for Javascript conversion.
        :param value: `str` The value to clean.
        :returns: `str` The cleaned value.
        """
        candidate = json.dumps(value)[1:-1]

        escape_chars = [("\\", "\\\\"), ('"', '\\"'), ("\n", "\\n")]
        for escape in escape_chars:
            candidate = candidate.replace(escape[0], escape[1])

        return candidate

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXEvent object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        clean_id = CXEvent.__clean_string(self.id)
        clean_script = CXEvent.__clean_string(self.script)
        rep_candidate = f"CXEvent(" f'id="{clean_id}", ' f'script="{clean_script}"' f")"
        return rep_candidate



"""
This tutorial is based on Plotly's Dash tutorial series, which begins at:
https://dash.plotly.com/installation

The intention of this tutorial is to demonstrate the basic CXDashElement class as a replacement for Plotly's own charts
but retaining the same overall application functionality.

CXDashElement is the essential means by which a CanvasXpress Dash element is added to a user interface.  It accepts
prepared config and event JSON strings to format and enhance a CanvasXpress chart.  CXDashElement is part of the
cxdash package that is bundled with the canvasxpress[dash] and canvasxpress[all] installations.
"""

# Run this app with `python app.py` and
# visit http://127.0.0.1:8050/ in your web browser.

import json
from dash import Dash, html
import cxdash

g_app = Dash(__name__)

# CanvasXpress chart configuration
_g_cx_data = {
    "y": {
        "smps": ["Apples", "Oranges", "Bananas"],
        "vars": ["SF", "Montreal"],
        "data": [
            [4, 1, 2],
            [2, 4, 5],
        ],
    },
}

_g_cx_colors = {
    "background": "#111111",
    "text": "rgb(127,219,255)",
    "bars": ["rgb(99,110,250)", "rgb(239,85,59)"],
}

_g_cx_config: dict = {
    "colors": _g_cx_colors["bars"],
    "graphOrientation": "vertical",
    "smpLabelRotate": 90,
    "plotBox": True,
    "plotBoxColor": "White",
    "background": _g_cx_colors["background"],
    "theme": "Plotly",
    "xAxis": ["Fruit", "City"],
    "xAxisTitle": "Amount",
    "xAxis2Show": False,
    "xAxisMinorTicks": False,
    "axisTitleColor": _g_cx_colors["text"],
    "axisTickColor": _g_cx_colors["text"],
    "legendColor": _g_cx_colors["text"],
    "smpTitle": "Fruit",
    "smpTitleFontColor": _g_cx_colors["text"],
    "smpLabelFontColor": _g_cx_colors["text"],
}

# Application
g_app.layout = html.Div(
    style={"backgroundColor": _g_cx_colors["background"]},
    children=[
        html.H1(
            children="Hello Dash",
            style={"textAlign": "center", "color": _g_cx_colors["text"]},
        ),
        html.H2(
            children="An Example of the Basic CXDashElement Class for Plotting a CanvasXpress Chart",
            style={"textAlign": "center", "color": _g_cx_colors["text"]},
        ),
        html.Div(
            id="chart-container",
            children=[
                html.Div(
                    id="cx-container",
                    style={"textAlign": "center"},
                    children=[
                        cxdash.CXDashElement(
                            id="example-cx",
                            data=json.dumps(_g_cx_data),
                            config=json.dumps(_g_cx_config),
                            width="600",
                            height="400",
                        ),
                    ],
                ),
            ],
        ),
    ],
)

if __name__ == "__main__":
    g_app.run_server(debug=False)



from abc import abstractmethod, ABC
from typing import Union

from canvasxpress.data.convert import CXDictConvertable

Y = "y"
X = "x"
Z = "z"
VENN = "venn"

VARS = "vars"
SMPS = "smps"
DATA = "data"
CORS = "cors"
LEGEND = "legend"
NODES = "nodes"


class CXData(CXDictConvertable):
    """
    CXData defines an essential data class for managing data acquisiton,
    transformation, and introspection as required by the `CanvasXPress` class.
    """

    @property
    @abstractmethod
    def data(self) -> dict:
        """
        A property accessor for the data managed by the object.  Regardless of
        the input data the returned data structure will be a dict-type for use
        with CanvasXpress.
        :returns: `dict`
            A dictionary representing a data map suitable for use with a chart.
        """
        pass

    @abstractmethod
    def get_raw_dict_form(self) -> dict:
        """
        Provides a simple dict perspective of the data with no metadata or other
        contextual transformations performed.  For example, if the data is
        natively in `dict` form then it would be passed-through with no
        modification or enhancement.
        :returns: `dict`
            The `dict` perspective of the data with as little modification or
            interpretation as is reasonable.
        """
        pass

    @abstractmethod
    def __init__(self, data: Union[object, None]) -> None:
        """
        Initializes the CXData object with data.
        :param data: `Union[object, None]`
            Given an object or no data prepares a new CXData instance ready for
            use by a `CanvasXpress` object.
        """
        pass


class CXDataProfileException(ValueError):
    """
    CXDataProfileException is used by CXDataProfile to describe validation
    errors when considering CXData objects in the context of CXDataProfile
    objects.
    """

    pass


class CXKeyPairData(CXData):
    """
    CXKeyPairData is a marker class to indicate that the data managed will be
    generally of the structure key-pair.  A `dict` is an example of key-pair
    data.
    """

    pass


class CXMatrixData(CXData):
    """
    CXMatrixData is a marker class to indicate that the data management will be
    generally of the structure matrix or tabular.  A spreadsheet is an example
    of matrix data.
    """

    pass



import json
from typing import Any, Union, List

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.base import CXRenderable

JSON_TEMPLATE: str = """
{
    "renderTo": "@renderTo@",
    "data": @data@,
    "config": @config@,
    "afterRender": @afterRender@,
    "otherParams": @otherParams@,
    "events": @events@,
    "width": @width@,
    "height": @height@
}
""".strip()


class CXJSON(CXRenderable):
    """
    CXJSON is a `CXRenderable` that renders `CanvasXpress` objects into reproducible JSON descriptions.
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXJSON` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    @classmethod
    def render_to_json(cls, cx: CanvasXpress):
        """
        Renders the given CanvasXpress object into a reproducible JSON form.
        :params cx: `CanasXpress`
            The `CanvasXpress` object to be converted.
        :returns: `str`
            A `str` bearing the reproducible JSON.
        """

        if cx is None:
            raise ValueError("cx cannot be None.")

        cx_element_params = {
            "renderTo": cx.render_to,
            "data": cx.data.render_to_dict(config=cx.config),
            "config": cx.config.render_to_dict(),
            "afterRender": cx.after_render.render_to_list(),
            "otherParams": cx.other_init_params.render_to_dict(),
            "events": "js_events",
            "width": cx.width,
            "height": cx.height,
        }

        # Support unique data without JSON data structure
        if cx_element_params["data"].get("raw"):
            cx_element_params["data"] = cx_element_params["data"]["raw"]

        cx_element_params["events"] = cx.events.render_to_js()

        reproducible_json = JSON_TEMPLATE
        for key in cx_element_params.keys():
            if isinstance(cx_element_params[key], str):
                value = cx_element_params[key]
            else:
                value = json.dumps(cx_element_params[key])
            reproducible_json = reproducible_json.replace(f"@{key}@", value)

        return reproducible_json

    def render(self, **kwargs: Any) -> list:
        """
        Renders the associated CanvasXpress object appropriate to create reproducible JSON.
        :param kwargs: `Any`
            * Unused.
        :returns: `list[str]`
            A list of `str`, one each of a reproducible JSON representation per tracked CanvasXpress object.
        """
        if self.canvas is None:
            return []
        elif isinstance(self.canvas, list):
            reproducible_jsons: list = []
            for cx in self.canvas:
                reproducible_jsons.append(CXJSON.render_to_json(cx))
            return reproducible_jsons
        else:
            return [CXJSON.render_to_json(self.canvas)]



"""
This tutorial is based on Plotly's Dash tutorial series, which begins at:
https://dash.plotly.com/installation

The intention of this tutorial is to demonstrate the CanvasXpress class as a replacement for Plotly's own charts
but retaining the same overall application functionality.

CXDashElement is the essential means by which a CanvasXpress Dash element is added to a user interface.  It accepts
prepared config and event JSON strings to format and enhance a CanvasXpress chart.  CXDashElement is part of the
cxdash package that is bundled with the canvasxpress[dash] and canvasxpress[all] installations.

CanvasXpress is a full-featured component for customizing charts: it eases working with data at the Python tier, and
it simplifies dynamic configuration.  Coupled with supporting components, such as CXDashElementFactory, it makes the
development of dynamic Dash, Flask, Django, Jupyter, and similar solutions easier.
"""

# Run this app with `python app.py` and
# visit http://127.0.0.1:8050/ in your web browser.=

from dash import Dash, html
from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.dash import CXElementFactory

g_app = Dash(__name__)

# No need to create_element many forms of data into CanvasXpress data objects; for example, use this CSV as-is.
_g_csv = """, Apples, Oranges, Bananas
SF      ,      4,       1,       2
Montreal,      2,       4,       5
"""

_g_cx_colors = {
    "background": "#111111",
    "text": "rgb(127,219,255)",
    "bars": ["rgb(99,110,250)", "rgb(239,85,59)"],
}

# Work with data as normal Python values.  cx_data could also have been a DataFrame, JSON, URL, etc.
_g_cx_chart = CanvasXpress(
    render_to="fruit_consumption",
    data=_g_csv,
    config={
        "colors": _g_cx_colors["bars"],
        "graphOrientation": "vertical",
        "smpLabelRotate": 90,
        "plotBox": True,
        "plotBoxColor": "White",
        "background": _g_cx_colors["background"],
        "theme": "Plotly",
        "xAxis": ["Fruit", "City"],
        "xAxisTitle": "Amount",
        "xAxis2Show": False,
        "xAxisMinorTicks": False,
        "axisTitleColor": _g_cx_colors["text"],
        "axisTickColor": _g_cx_colors["text"],
        "legendColor": _g_cx_colors["text"],
        "smpTitle": "Fruit",
        "smpTitleFontColor": _g_cx_colors["text"],
        "smpLabelFontColor": _g_cx_colors["text"],
    },
    width=650,
    height=450,
)

# Application
g_app.layout = html.Div(
    style={"backgroundColor": _g_cx_colors["background"]},
    children=[
        html.H1(
            children="Hello Dash",
            style={"textAlign": "center", "color": _g_cx_colors["text"]},
        ),
        html.H2(
            children=(
                "An Example of the Advanced CanvasXpress and CXDashElementFactory"
                " Classes for Plotting a CanvasXpress Chart"
            ),
            style={"textAlign": "center", "color": _g_cx_colors["text"]},
        ),
        html.Div(
            id="chart-container",
            children=[
                html.Div(
                    id="cx-container",
                    style={"textAlign": "center"},
                    children=CXElementFactory.render(_g_cx_chart),
                ),
            ],
        ),
    ],
)

if __name__ == "__main__":
    g_app.run_server(debug=False)



from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.popup import CXBrowserPopup

import pandas as pd

# Each file is a tab-delimited matrix with columns and an index, and we can
# use pandas to read each at their URL to create dataframes

y_data_url = "https://www.canvasxpress.org/data/cX-scatterR2-dat.txt"
y_data = pd.read_csv(y_data_url, sep="\t", index_col=0)

z_data_url = "https://www.canvasxpress.org/data/cX-scatterR2-var.txt"
z_data = pd.read_csv(z_data_url, sep="\t", index_col=0)

# The dict structure is an easy way to shape and submit data to CanvasXpress
volcano_data = {
    # Data
    "y": {
        # Samples, which can be considered the columns
        "smps": y_data.columns.to_list(),
        # Variables, which can be considered the index
        "vars": y_data.index.tolist(),
        # The data itself, split into rows per variable and ordered by samples
        "data": y_data.values.tolist(),
    },
    # Variable annotations
    "z": {col: z_data[col].to_list() for col in z_data.columns.to_list()},
}

chart = CanvasXpress(
    data=volcano_data,
    config={
        "graphType": "Scatter2D",
        "title": "Volcano plot",
        "axisAlgorithm": "rPretty",
        "backgroundType": "window",
        "backgroundWindow": "rgb(238,238,238)",
        "colorBy": "Group",
        "colors": ["rgba(0,104,139,0.5)", "rgba(205,0,0,0.5)", "rgba(64,64,64,0.5)"],
        "decorations": {
            "line": [
                {"color": "rgba(205,0,0,0.5)", "width": 2, "x": 0.5},
                {"x": -0.5, "width": 2, "color": "rgba(0,104,139,0.5)"},
            ]
        },
        "legendBackgroundColor": "rgb(238,238,238)",
        "legendBox": True,
        "legendBoxColor": "rgb(0,0,0)",
        "plotBox": False,
        "showDecorations": True,
        "showTransition": False,
        "sizeBy": "FC",
        "sizes": [4, 14, 16, 18],
        "theme": "CanvasXpress",
        "xAxis": ["logFC"],
        "xAxisTickColor": "rgb(255,255,255)",
        "yAxis": ["-log-pVal"],
        "yAxisTickColor": "rgb(255,255,255)",
    },
    width=613,
    height=613,
)

browser = CXBrowserPopup(chart)
browser.render()



from abc import ABC, abstractmethod
from copy import copy
from typing import Any, List, Union

from canvasxpress.canvas import CanvasXpress


class CXRenderAssociation(ABC):
    """
    CXRenderAssociation tracks A set of CanvasXpress objects that will be used to produce
    a domain- or container-specific rendering (e.g., for display in a Jupyter notebook).
    """

    __cx: List[CanvasXpress] = list()
    """
    The CanvasXpress object to be managed by this CXRenderAssociation.
    """

    @property
    def canvas(self) -> Union[List[CanvasXpress], CanvasXpress, None]:
        """
        Provides the tracked CanvasXpress object.
        :returns: `value: Union[List[CanvasXpress], CanvasXpress, None]`
            A `list` of `CanvasXpress` objects if multiple are tracked,
            or one `CanvasXpress object, or `None` if no objects are tracked.
        """
        if len(self.__cx) == 0:
            return None

        elif len(self.__cx) == 1:
            return self.__cx[0]

        else:
            return copy(self.__cx)

    @canvas.setter
    def canvas(self, value: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Sets the CanvasXpress object to be tracked.
        :praram value: `value: Union[List[CanvasXpress], CanvasXpress, None]`
            A `list` of `CanvasXpress` objects if multiple are tracked,
            or one `CanvasXpress object, or `None` if no objects are to be
            tracked.
        """
        if value is None:
            self.__cx = list()

        elif isinstance(value, CanvasXpress):
            self.__cx = [value]

        elif isinstance(value, list):
            for v in value:
                if not isinstance(v, CanvasXpress):
                    raise TypeError("value must of type CanvasXpress")

            self.__cx = copy(value)

        else:
            raise TypeError("value must of type CanvasXpress")

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXRenderable` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        charts = list()
        for arg in cx:
            if arg is None:
                pass

            if isinstance(arg, (list, tuple)):
                for item in arg:
                    if not isinstance(item, CanvasXpress):
                        raise TypeError("All cx members must be of type CanvasXpress")
                    else:
                        charts.append(item)

            elif isinstance(arg, CanvasXpress):
                charts.append(arg)

            else:
                raise TypeError("All cx members must be of type CanvasXpress")

        self.canvas = charts


class CXRenderable(CXRenderAssociation):
    """
    CXRenderable is capable of rendering a CanvasXpress object to some kind of
    output or display device.
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXRenderable` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    @abstractmethod
    def render(self, **kwargs: Any):
        """
        Renders the associated CanvasXpress object appropriate to the render_to.
        Not implemented.
        :param kwargs: `Any`
            Parameters specific to implementations are supported.  The essential
            render call should work with no extra parameters, and with
            parameters that do not apply to the implementation.
        """
        pass


class CXRenderFactory(CXRenderAssociation):
    """
    CXRenderFactory produces objects for use in a container or framework that understand how
    to cooperate with the framework to produce CanvasXpress illustrations.
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXRenderFactory` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    @abstractmethod
    def render_all(self, **kwargs: Any) -> List[object]:
        """
        Provides a list of objects that can be used by the target domain or container
        to create CanvasXpress illustrations or instantiations.
        Not implemented.
        :param kwargs: `Any`
            Parameters specific to implementations are supported.  The essential
            render call should work with no extra parameters, and with
            parameters that do not apply to the implementation.
        """
        pass




from setuptools import setup, find_packages

long_description = '''# CanvasXpress for Python

<a href="https://www.canvasxpress.org">
<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/images/hexagon.png" align="left" width="175" style="vertical-align:middle;margin:10px 10px"> 
</a>
<br>

[***CanvasXpress***](https://www.canvasxpress.org) was developed as the core visualization component for bioinformatics
and systems biology analysis at Bristol-Myers Squibb. It supports a large number
of [visualizations ](https://www.canvasxpress.org/examples.html)
to display scientific and non-scientific data. ***CanvasXpress*** also includes a simple and unobtrusive
[user interface](https://www.canvasxpress.org/docs/interface.html) to explore complex data sets, a sophisticated and
unique mechanism to keep track of all user customization for
[Reproducible Research ](https://www.canvasxpress.org/docs/audit.html) purposes, as well as an 'out of the box'
broadcasting capability to synchronize selected data points across all ***CanvasXpress*** plots in a page. Data can be
easily sorted, grouped, transposed, transformed or clustered dynamically. The fully customizable mouse events as well as
the zooming, panning and drag-and-drop capabilities are features that make this library unique in its class.

<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/images/sample_graphs.png" align="center" width="726"></a>

***CanvasXpress*** can be used for native Python integration for the following environments:

- [Shiny for Python](https://shiny.posit.co/py/)
- [Streamlit](https://streamlit.io/)
- [Plotly Dash](https://dash.plotly.com/)
- [Jupyter](https://jupyter.org/)
- [Flask](https://flask.palletsprojects.com/en/1.1.x/)
- [Django](https://www.djangoproject.com/)

The RStudio IDE Viewer is also used when running code chunks in Jupyter, Quarto, and RMD Python code chunks.

This ***CanvasXpress*** Python package is maintained by [Dr. Todd C. Brett](https://github.com/docinfosci), with support
from [Aggregate Genius Inc.](https://www.aggregate-genius.com), in cooperation with Dr. Isaac Neuhaus and the
***CanvasXpress*** team.

## Project Status

[![Release](https://img.shields.io/pypi/v/canvasxpress.svg)](https://pypi.org/project/canvasxpress)
[![Compatibility](https://img.shields.io/pypi/pyversions/canvasxpress.svg)](https://pypi.org/project/canvasxpress)
[![Implementations](https://img.shields.io/pypi/implementation/canvasxpress.svg)](https://pypi.org/project/canvasxpress)
[![PyPI - Downloads](https://img.shields.io/pypi/dm/canvasxpress)](https://pypi.org/project/canvasxpress)
[![docinfosci](https://circleci.com/gh/docinfosci/canvasxpress-python/tree/main.svg?style=shield)](https://circleci.com/gh/docinfosci/canvasxpress-python/?branch=main)
[![Coverage Status](https://coveralls.io/repos/github/docinfosci/canvasxpress-python/badge.svg?branch=main)](https://coveralls.io/github/docinfosci/canvasxpress-python?branch=main)
[![Activity](https://img.shields.io/github/last-commit/docinfosci/canvasxpress-python/develop)](https://github.com/docinfosci/canvasxpress-python)

## Documentation and Installation

### Documentation

Documentation is maintained on this page and at [CanvasXpress.org](https://www.canvasxpress.org) and LinkedIn:

- [Introduction to CanvasXpress for Python](https://www.linkedin.com/pulse/introducing-canvasxpress-python-todd-brett-hew0f/?trackingId=G8kTE2QyRH%2BrcVSzxJc8Hg%3D%3D)

### Installation

CanvasXpress for Python can be installed using _pip_:

The essential CanvasXpress package, for use with the CLI or flask and similar environments, can be installed using:

```terminal
pip install canvasxpress
```

or

```terminal
pip install "canvasxpress[core]"
```

In addition to _core_, the following additional targets can be used:

- _jupyter_ - installs additional packages to support rendering in Jupyter, Quarto, and IPython documents
- _dash_ - installs additional packages to support rendering in Plotly Dash applications
- _streamlit_ - installs additional packages to support rendering in Snowflake Streamlit applications
- _shiny_ - installs additional packages to support rendering in Posit Shiny for Python applications
- _rstudio_ - installs additional packages to support rendering in the Posit RStudio IDE Viewer, plus includes the same
  packages for jupyter and shiny
- _all_ - installs all additional packages to support rendering in any supported document or application

## Usage

This section provides general instructions on creating and customizing charts.

<details>
<summary>Click to read more</summary>

### Drawing Charts

The `CanvasXpress` object defines what a chart should contain and how it should be formatted, but rendering the chart is
performed by the functions `graph()` and `show_in_browser()`.

To use `graph()` import it from `canvasxpress.plot` and then call it by passing a `CanvasXpress` object. For example, a
Quarto, RMD, or Jupyter Notebook code chunk could be:

```python
from random import random

from canvasxpress.canvas import CanvasXpress
from canvasxpress.plot import graph

graph(
    CanvasXpress(
        data={
            "y": {
                "data": [
                    [random() % 100 for i in range(20)]
                ],
                "vars": ["A"],
            }
        },
        config={
            "background": "rgb(255,255,255)",
            "colorScheme": "CanvasXpress",
            "graphOrientation": "vertical",
            "graphType": "Area",
            "objectBorderColor": False,
            "plotBox": False,
            "plotBoxColor": "rgb(204,204,204)",
            "showLegend": False,
            "showLegendBorder": True,
            "smpLabelRotate": 90,
            "smpTitle": "time",
            "xAxis": ["A"],
            "xAxisTickRightShow": False,
            "yAxisTickTopShow": False
        },
        width=609,
        height=609
    )
)
```

Some application frameworks, such as _Shiny for Python_ and _Plotly Dash_, expect an object to be rendered to the
framework as part of the reactive flow. In these contexts, the `graph()` function creates an appropriate object and
returns it. That value can be assigned to a variable to be returned at a later point in the code, or be immediately
returned. See the Shiny for Python and Dash examples for specific usage.

`show_in_browser()` is similar to `graph()` except that it opens a browser window on the local system and displays the
chart. It's used to facilitate learning and debugging.

`graph()` does a good job of determining the runtime context to choose how the chart should be rendered, but in the case
installed packages or runtime configurations confuse the function an environment variable can be set to override
how `graph()` performs the rendering. Set `CANVASXPRESS_TARGET_CONTEXT` to be one of these values as appropriate for
this situation (and don't forget to pip install the necessary package support):

- rstudio
- shiny
- jupyter
- dash
- streamlit
- browser

For example:

```python
from os import environ

environ["CANVASXPRESS_TARGET_CONTEXT"] = "jupyter"
```

or via a shell (_bash_ example provided):

```shell
export CANVASXPRESS_TARGET_CONTEXT="jupyter"
```

### Default and Pinned CanvasXpress JavaScript Editions

CanvasXpress for Python generates JavaScript that assumes use of the latest available edition of CanvasXpress for
JavaScript, but it can be set to use a specific edition.

Review this site for available versions:
https://cdnjs.com/libraries/canvasXpress

The desired version is expressed as a `str`. Prior to generating a CanvasXpress chart use the following code to set the
edition that shall be used:

```python
from canvasxpress.canvas import CanvasXpress

CanvasXpress.set_cdn_edition("48.3")  # Or whatever available version is desired.
```

To use the default edition once again during the runtime session set the value to `None`.

This is the best way to assure a specific chart behavior for production application releases; however, once set any new
JavaScript edition features or fixes will not be available until the code is removed or a different version is set.

Similarly, a custom URL for the JavaScript or CSS libraries can be set as well (for example, to facilitate development
of CanvasXpress JS).

```python
from canvasxpress.canvas import CanvasXpress

CanvasXpress.set_js_library_url("http://localhost:8080/js")
CanvasXpress.set_css_library_url("http://localhost:8080/css")
```

### Customizing Charts

Generally speaking, a `CanvasXpress` object accepts the following parameters:

#### render_to

`render_to` is a `str` value that identifies the chart when rendered into HTML. JavaScript functions can use this ID to
access the chart and perform CanvasXpress operations within the browser. Omitting `render_to` or setting it to
`None` will make the `CanvasXpress` object assume an anonymous mode in which a new GUID will be generated each time
`graph()` is called. If the chart will not be maniluated using JavaScript in the browser it is fine for charts to be
anonymous.

__NOTE:__ React environments regularly destroy and rebuild objects as the page is updated. In these environments it is
possible for the timing of object destruction and JavaScript execution to cause a crash. The best defense is to either
use anonymous mode, or if an ID must be known then a unique identifier should be set each time `graph()` is called. In
this manner an ID for a chart in the middle of being recreated is never referenced. For example:

```python
chart = CanvasXpress(...)
chart.render_to = str(guid4()).replace("-", "_")
return graph(chart)
```

Plotly's Dash framework uses React, and Dash applications should consider using only anonymous charts or assigning
unique values as the ID similar to the above code. Shiny for Python does not seem to suffer from this challenge.

#### data

`data` sets the chart's data and metadata. This is an involved topic, and
the [introductory article](https://www.linkedin.com/pulse/introducing-canvasxpress-python-todd-brett-hew0f/?trackingId=G8kTE2QyRH%2BrcVSzxJc8Hg%3D%3D)
is an excellent read to understand how data should be shaped. In general, data will be a `dict` (also known as an XYZ
object), Web URL, or `str`.

Data dict example:

```python
data_for_use_in_chart = {
    "y": {
        "data": [
            [random() % 100 for i in range(20)]
        ],
        "vars": ["A"],
    }
}
```

CanvasXpress for Python supports Pandas DataFrame assignments to `data`, plus supporting `sample_annotation` and
`variable_annotation` properties. If data is assigned in this manner, then on generation of the JavaScript the
CanvasXpress object will create an XYZ object using the DataFrame(s).

```python
from canvasxpress.canvas import CanvasXpress
import pandas
import io
import requests

data_url = "https://www.canvasxpress.org/data/cX-generic-dat.txt"
data_raw = requests.get(data_url).content
data_df = pandas.read_csv(
    io.StringIO(data_raw.decode('utf-8')),
    sep="\t",
    index_col=0,
)

sample_annot_url = "https://www.canvasxpress.org/data/cX-generic-smp.txt"
sample_annot_raw = requests.get(sample_annot_url).content
sample_annot_df = pandas.read_csv(
    io.StringIO(sample_annot_raw.decode('utf-8')),
    sep="\t",
    index_col=0,
)

variable_annot_url = "https://www.canvasxpress.org/data/cX-generic-var.txt"
variable_annot_raw = requests.get(variable_annot_url).content
variable_annot_df = pandas.read_csv(
    io.StringIO(variable_annot_raw.decode('utf-8')),
    sep="\t",
    index_col=0,
)

cx = CanvasXpress(
    data=data_df,
    sample_annotation=sample_annot_df,
    variable_annotation=variable_annot_df,
    config={
        "graphOrientation": "vertical",
        "graphType": "Dotplot",
        "lineType": "spline",
        "llmHeader": [
            ["V1", "V2", "V3", "V4", "Factor1", "Factor2", "Factor3", "Factor4", "Factor5", "Factor6"]
        ],
        "showSmpOverlaysLegend": True,
        "smpOverlayProperties": {
            "Factor4": {
                "color": "blue",
                "thickness": 50,
                "type": "Bar",
                "showLegend": "True"
            },
            "Factor5": {
                "thickness": 50,
                "color": "grey",
                "type": "Bar",
                "showLegend": "True"
            },
            "Factor6": {
                "showLegend": "True",
                "thickness": 50,
                "color": "red",
                "type": "Bar"
            },
            "Factor1": {
                "type": "Default",
                "color": "rgb(10,176,219)",
                "spectrum": ["rgb(69,117,180)", "rgb(145,191,219)", "rgb(224,243,248)", "rgb(255,255,191)",
                             "rgb(254,224,144)", "rgb(252,141,89)", "rgb(215,48,39)"],
                "scheme": "CanvasXpress",
                "hideName": False,
                "hideValue": False,
                "showLegend": True,
                "legendColumns": False,
                "showBox": True,
                "ticksOnBottomOrLeft": True,
                "rotate": False,
                "invert": False,
                "position": "bottom"
            },
            "Factor2": {
                "type": "Default",
                "color": "rgb(254,211,133)",
                "spectrum": ["rgb(69,117,180)", "rgb(145,191,219)", "rgb(224,243,248)", "rgb(255,255,191)",
                             "rgb(254,224,144)", "rgb(252,141,89)", "rgb(215,48,39)"],
                "scheme": "CanvasXpress",
                "hideName": False,
                "hideValue": False,
                "showLegend": True,
                "legendColumns": False,
                "showBox": True,
                "ticksOnBottomOrLeft": True,
                "rotate": False,
                "invert": False,
                "position": "bottom"
            },
            "Factor3": {
                "type": "Default",
                "color": "rgb(254,105,105)",
                "spectrum": ["rgb(69,117,180)", "rgb(145,191,219)", "rgb(224,243,248)", "rgb(255,255,191)",
                             "rgb(254,224,144)", "rgb(252,141,89)", "rgb(215,48,39)"],
                "scheme": "CanvasXpress",
                "hideName": False,
                "hideValue": False,
                "showLegend": True,
                "legendColumns": False,
                "showBox": True,
                "ticksOnBottomOrLeft": True,
                "rotate": False,
                "invert": False,
                "position": "bottom"
            }
        },
        "smpOverlays": ["Factor1", "Factor2", "Factor3"],
        "smpTextRotate": 45,
        "smpTitle": "Collection of Samples",
        "smpTitleFontStyle": "italic",
        "subtitle": "Random Data",
        "theme": "CanvasXpress",
        "title": "Dotplot Graph",
        "xAxis": ["V1", "V2", "V3", "V4"],
        "xAxisTickFormat": "%.0f Mil."
    },
)
```

If a DataFrame is assigned to `data` then its index will be used for `xyz["y"]["vars"]` values and its header will be
used for the `xyz["y"]["smps"]` values.

If a DataFrame is assigned to `sample_annotation` or `variable_annotation` then the following strategy will be used to
determine how the DataFrame should be parsed to create the corresponding `x` and `z` properties, respectively:

1. The first column is first inspected to see if all of the column values match one of the available
   sample (`xyz["y"]["smps"]` aka columns) or variable (`xyz["y"]["vars"]` aka rows) values.
2. If a match is not made within the column, then the first row is inspected for the same criteria.
3. Next, the DataFrame headers will be inspected.
4. Finally, the DataFrame index will be used regardless of a match.

If no DataFrame is assigned to the `sample_annotation` or `variable_annotation` properties then those portions of the
xyz object will simply be ommitted.

`sample_annotation` and `variable_annotation` can only be used if the `data` property is already a DataFrame. If any
other type is used, such as a `dict` or `str` then an exception will be raised.

`data`, `sample_annotation`, and `variable_annotation` can be set to `None` to remove their values.

Data URL example:

```python
data_for_use_in_chart = "https://corgis-edu.github.io/corgis/datasets/csv/state_demographics/state_demographics.csv"
```

Data text (CSV) example:

```python
data_for_use_in_chart = """
"State","Population.Population Percent Change","Population.2014 Population"
"Connecticut","-10.2","3605944"
"Delaware","8.4","989948"
"""
```

#### config

`config` describes the chart's formatting. It is a `dict` in which properties are specified and assigned values. All of
the values must be compliant with Python's `json.dumps()` function. For example:

```python
config = {
             "background": "rgb(255,255,255)",
             "colorScheme": "CanvasXpress",
             "graphOrientation": "vertical",
             "graphType": "Area",
             "objectBorderColor": False,
             "plotBox": False,
             "plotBoxColor": "rgb(204,204,204)",
             "showLegend": False,
             "showLegendBorder": True,
             "smpLabelRotate": 90,
             "smpTitle": "time",
             "xAxis": ["A"],
             "xAxisTickRightShow": False,
             "yAxisTickTopShow": False
         },
```

Configuration options may also be specified as keyword arguments in the `CanvasXpress` function call. For example:

```python
CanvasXpress(
    data={
        "y": {
            "vars": ["Gene1"],
            "smps": ["Smp1", "Smp2", "Smp3"],
            "data": [[10, 35, 88]]
        }
    },
    graphOrientation="vertical",
    graphType="Bar",
    showLegend=False,
    smpLabelRotate=90,
    smpTitle="Samples",
    theme="CanvasXpress",
    title="Bar Graph Title",
    xAxisTitle="Value"
)
```

If `config` and configuration keyword arguments are specified together, their configuration options are combined, with
the keyword arguments overriding any item of the same name in `config`. This approach can be used to define a generic
configuration and allow for ad hoc modifications.

#### width and height

`width` and `height` specify the chart's dimensions as pixels. If ommitted the CanvasXpress edition active for the
browser will assign default values, such as 500px by 500px.

#### Javascript Events

CanvasXpress provides support for Javascript events via hook functions that are called when events occur, such as mouse
movement or clicks. These events are supported via the canvasxpress.js sub-package. `CXEvent` objects hold the
Javascript instructions for Web events. An example event for graph clicks with popup information is:

```python
from canvasxpress.js.function import CXEvent

CXEvent(
    id="click",
    script="""
    var s = 'click on var ' + o.y.vars[0] + ' and smp ' + o.y.smps[0];
    t.showInfoSpan(e, s);
    """
)
```

The general JavaScript template of a CanvasXpress Javascript hook function is:

```javascript
function (o, e, t) {
    // script logic goes here
};
```

`CXEvent` objects can be provided as a single object or as a list. Here's an example of an event the provides additional
information about chart data upon a user click:

```python
from canvasxpress.canvas import CanvasXpress
from canvasxpress.plot import graph
from canvasxpress.js.function import CXEvent

graph(
    CanvasXpress(
        render_to="example_chart",
        data={
            "y": {
                "vars": ["Gene1"],
                "smps": ["Smp1", "Smp2", "Smp3"],
                "data": [[10, 35, 88]]
            }
        },
        config={
            "graphOrientation": "vertical",
            "graphType": "Bar",
            "showLegend": False,
            "smpLabelRotate": 90,
            "smpTitle": "Samples",
            "theme": "CanvasXpress",
            "title": "Bar Graph Title",
            "xAxisTitle": "Value"
        },
        events=[
            CXEvent(
                id="click",
                script="""
                var s = 'click on var ' + o.y.vars[0] + ' and smp ' + o.y.smps[0];
                t.showInfoSpan(e, s);
                """
            ),
        ]
    )
)
```

### Converting to and from Reproducible JSON

CanvasXpress for Python can also convert to and from reproducible JSONs usable with the JavaScript and R editions of the
library.  `convert_to_reproducible_json` takes an existing CanvasXpress object and provides a `str` copy of the JSON,
which can then be logged for debugging or saved to disk for use elsewhere.  `convert_from_reproducible_json` does the
opposite by taking a reproducible JSON `str` and providing the CanvasXpress object equivalent.

_Note: Events are not currently supported for import. This will be provided in a future edition. Export supports
events._

For example, do the following to see the JSON in the Python console:

```python
from canvasxpress.canvas import CanvasXpress
from canvasxpress.plot import convert_to_reproducible_json
from canvasxpress.js.function import CXEvent

print(
    convert_to_reproducible_json(
        CanvasXpress(
            render_to="example_chart",
            data={
                "y": {
                    "vars": ["Gene1"],
                    "smps": ["Smp1", "Smp2", "Smp3"],
                    "data": [[10, 35, 88]]
                }
            },
            config={
                "graphOrientation": "vertical",
                "graphType": "Bar",
                "showLegend": False,
                "smpLabelRotate": 90,
                "smpTitle": "Samples",
                "theme": "CanvasXpress",
                "title": "Bar Graph Title",
                "xAxisTitle": "Value"
            },
            events=[
                CXEvent(
                    id="click",
                    script="""
                    var s = 'click on var ' + o.y.vars[0] + ' and smp ' + o.y.smps[0];
                    t.showInfoSpan(e, s);
                    """
                ),
            ]
        )
    )
)
```

The console would display:

```text
{
    "renderTo": "example_chart",
    "data": {"y": {"vars": ["Gene1"], "smps": ["Smp1", "Smp2", "Smp3"], "data": [[10, 35, 88]]}, "x": {}, "z": {}},
    "config": {"graphOrientation": "vertical", "graphType": "Bar", "showLegend": false, "smpLabelRotate": 90, "smpTitle": "Samples", "theme": "CanvasXpress", "title": "Bar Graph Title", "xAxisTitle": "Value"},
    "afterRender": [],
    "otherParams": {},
    "events": {'click': function(o, e, t){
                    var s = 'click on var ' + o.y.vars[0] + ' and smp ' + o.y.smps[0];
                    t.showInfoSpan(e, s);
                    }},
    "width": 500,
    "height": 500
}
```

This text could be saved to a file, such as `example.json`, and then dragged onto a CanvasXpress chart in a browser to
load the equivalent chart. In fact, CanvasXpress for Python uses the core functionality producing JSON output to make
charts available in contexts such as Dash and Shiny.
</details>

## Application, NoteBook, and Console Examples

### Rendering Charts in the RStudio IDE Viewer Pane

The RStudio IDE's Viewer panel is now supported for rendering interactive charts in the Viewer!  When the `graph()`
function is called it detects that RStudio is running and renders the chart in the Viewer instead of a document, such as
for Quarto code chunks. However, if the document is a Quarto or RMD file and the appropriate HTML (etc.)
generation is performed then the CanvasXpress charts will be embedded in the generated output file as normal.

### A Basic Python Script / Console Example

Charts can be defined in scripts or a console session and then displayed using the default browser, assuming that a
graphical browser with Javascript support is available on the host system. To do so use the `show_in_browser()`
function instead of `graph()`.

<details>
<summary>Click to read more</summary>

```python
from canvasxpress.canvas import CanvasXpress
from canvasxpress.plot import show_in_browser

if __name__ == "__main__":
    # Define a CX bar chart with some basic data
    chart: CanvasXpress = CanvasXpress(
        data={
            "y": {
                "vars": ["Gene1"],
                "smps": ["Smp1", "Smp2", "Smp3"],
                "data": [[10, 35, 88]]
            }
        },
        config={
            "graphType": "Bar"
        }
    )

    # Display the chart in its own Web page
    show_in_browser(chart)
```

Upon running the example the following chart will be displayed on systems such as MacOS X, Windows, and Linux with
graphical systems:

<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/examples/flask_bar_chart_basic.png" align="center" width="600"></a>
</details>

### A Shiny for Python Example

[Shiny for Python](https://shiny.posit.co/py/) is a new dashboard framework inspired by the highly successful Shiny for
R framework produced by Posit (formerly RStudio). This example shows how to create a basic Shiny for Python application
using a CanvasXpress Shiny element.

<details>
<summary>Click to read more</summary>

A basic Shiny for Python app provides a means by which:

1. A local development server can be started
1. A function can respond to input or draw an initial UI

First install Shiny for Python and CanvasXpress for Python:

```terminal
pip install shiny
pip install canvasxpress[shiny]
```

Then create a demo file, such as `app.py`, and insert:

```python
from random import random

from shiny import App, ui, render, reactive

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.shiny import output_canvasxpress
from canvasxpress.plot import graph

app_ui = ui.page_fluid(
    ui.row(
        ui.input_slider(
            "points_desired",
            "Points",
            min=0,
            max=100,
            value=0,
        ),
    ),
    ui.row(
        output_canvasxpress("chart_view"),
    )
)


def server(input, output, session):
    @render.ui
    @reactive.event(input.points_desired)
    def chart_view():
        return graph(
            CanvasXpress(
                data={
                    "y": {
                        "data": [
                            [random() % 100 for i in range(input.points_desired())]
                        ],
                        "vars": ["A"],
                    }
                },
                config={
                    "background": "rgb(255,255,255)",
                    "colorScheme": "CanvasXpress",
                    "graphOrientation": "vertical",
                    "graphType": "Area",
                    "objectBorderColor": False,
                    "plotBox": False,
                    "plotBoxColor": "rgb(204,204,204)",
                    "showLegend": False,
                    "showLegendBorder": True,
                    "smpLabelRotate": 90,
                    "smpTitle": "time",
                    "xAxis": ["A"],
                    "xAxisTickRightShow": False,
                    "yAxisTickTopShow": False
                },
                width=500,
                height=500
            )
        )


app = App(app_ui, server)
```

#### Run the App and View the Page

On the command line, execute:

```terminal
shiny run --reload --launch-browser app.py
```

And output similar to the following will be provided:

```terminal
Uvicorn running on http://127.0.0.1:8000 (Press CTRL+C to quit)
```

Browsing to `http://localhost:8000/` will result in a page with a CanvasXpress chart, which is being hosted by the Shiny
for Python framework:

<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/examples/shiny_chart_example.png" align="center" width="600"></a>

Congratulations!  You have created a Shiny for Python CanvasXpress app!
</details>

### A Streamlit Example

[Streamlit](https://streamlit.io) is a popular dashboard framework that is simplified compared to Dash and Shiny, but
just as powerful in terms of reactivity and extensions. This example shows how to create a basic Streamlit application
using a CanvasXpress Streamlit element.

<details>
<summary>Click to read more</summary>

A basic Streamlit app provides a means by which:

1. A local development server can be started
1. A function can respond to a URL

First install Streamlit and CanvasXpress for Python:

```terminal
pip install streamlit
pip install canvasxpress[streamlit]
```

Then create a demo file, such as `app.py`, and insert:

```python
import random

import streamlit as st

from canvasxpress.canvas import CanvasXpress
from canvasxpress.plot import graph

# A basic bar chart.  It's anonymous, so no render_to.  Data is added during the draw phase.
bar_chart = CanvasXpress(
    config={
        "graphOrientation": "vertical",
        "plotBox": True,
        "showLegend": False,
        "smpLabelRotate": 90,
        "smpTitle": "Samples",
        "theme": "CanvasXpress",
        "title": "Bar Graph Title",
        "xAxis": ["V1"],
        "xAxisTitle": "Value",
        "graphType": "Bar"
    },
    width=500,
    height=500
)

# Write the UI to the browser
# This code will be re-executed with each click of the button

# Name the theme
st.title('CanvasXpress in Streamlit!')

# Some columns to organize the button and chart
column1, column2 = st.columns([1, 3])

# A column with our data generator button
with column1:
    # This has no associated action, so by default it triggers a redraw of the UI.
    st.button("Generate New Data")

# Another column with the chart displayed
# With each redraw generate new random values
bar_chart.data = {
    "y": {
        "vars": ["V1"],
        "smps": ["S1", "S2", "S3"],
        "data": [
            [
                random.randint(100, 10000),
                random.randint(100, 10000),
                random.randint(100, 10000),
            ]
        ]
    }
}
with column2:
    # This plots the CanvasXpress chart into the UI.
    graph(bar_chart)
```

#### Run the App and View the Page

On the command line, execute:

```terminal
streamlit run app.py
```

And output similar to the following will be provided:

```terminal
Running on http://localhost:8501/ (Press CTRL+C to quit)
```

Browsing to `http://localhost:8501/` will result in a page with a CanvasXpress chart, which is being hosted by the
Streamlit framework:

<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/examples/streamlit_chart_basic.png" align="center" width="600"></a>

Congratulations!  You have created a Streamlit CanvasXpress app!
</details>

### A Dash Example

[Plotly Dash](https://dash.plotly.com/) is a popular dashboard framework similar to Shiny for Python or R. Dash
applications are Web pages with widgets and elements facilitating the interactive presentation of information. This
example shows how to create a basic Dash application using a CanvasXpress Dash element.

<details>
<summary>Click to read more</summary>

#### Create a Basic Dash App

A basic Dash app provides a means by which:

1. A local development server can be started
1. A function can respond to a URL

First install Dash and CanvasXpress for Python:

```terminal
pip install dash
pip install canvasxpress[dash]
```

Then create a demo file, such as `app.py`, and insert:

```python
from random import random

from dash import Dash, html

from canvasxpress.canvas import CanvasXpress
from canvasxpress.plot import graph

g_app = Dash(__name__)

colors = {
    "background": "#111111",
    "text": "rgb(127,219,255)",
}

# Application
g_app.layout = html.Div(
    style={"backgroundColor": colors["background"]},
    children=[
        html.H1(
            children="Hello Dash",
            style={"textAlign": "center", "color": colors["text"]},
        ),
        html.H2(
            children=(
                "An Example of the Advanced CanvasXpress and CXDashElementFactory"
                " Classes for Plotting a CanvasXpress Chart"
            ),
            style={"textAlign": "center", "color": colors["text"]},
        ),
        html.Div(
            id="chart-container",
            children=[
                html.Div(
                    id="cx-container",
                    style={"textAlign": "center"},
                    children=graph(
                        CanvasXpress(
                            data={
                                "y": {
                                    "data": [
                                        [random() % 100 for i in range(5)]
                                    ],
                                    "vars": ["A"],
                                }
                            },
                            config={
                                "background": "rgb(255,255,255)",
                                "colorScheme": "CanvasXpress",
                                "graphOrientation": "vertical",
                                "graphType": "Area",
                                "objectBorderColor": False,
                                "plotBox": False,
                                "plotBoxColor": "rgb(204,204,204)",
                                "showLegend": False,
                                "showLegendBorder": True,
                                "smpLabelRotate": 90,
                                "smpTitle": "time",
                                "xAxis": ["A"],
                                "xAxisTickRightShow": False,
                                "yAxisTickTopShow": False
                            },
                            width=500,
                            height=500
                        )
                    ),
                ),
            ],
        ),
    ],
)

if __name__ == "__main__":
    g_app.run_server(debug=True)
```

#### Run the App and View the Page

On the command line, execute:

```terminal
python3 app.py
```

And output similar to the following will be provided:

```terminal
Running on http://127.0.0.1:8050/ (Press CTRL+C to quit)
```

Browsing to `http://127.0.0.1:8050/` will result in a page with a CanvasXpress chart, which is being hosted by the Dash
framework:

<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/examples/dash_chart_basic.png" align="center" width="600"></a>

Congratulations!  You have created a Plotly Dash CanvasXpress app!
</details>

### A Flask Example

[Flask](https://palletsprojects.com/p/flask/) is a popular lean Web development framework for Python based applications.
Flask applications can serve Web pages, RESTful APIs, and similar backend service concepts. This example shows how to
create a basic Flask application that provides a basic Web page with a CanvasXpress chart composed using Python in the
backend.

The concepts in this example equally apply to other frameworks that can serve Web pages, such as Django and Tornado.

<details>
<summary>Click to read more</summary>

#### Create a Basic Flask App

A basic Flask app provides a means by which:

1. A local development server can be started
1. A function can respond to a URL

First install Flask and CanvasXpress for Python:

```terminal
pip install -U Flask canvasxpress
```

Then create a demo file, such as `app.py`, and insert:

```python
# save this as app.py
from flask import Flask

app = Flask(__name__)


@app.route('/')
def canvasxpress_example():
    return "Hello!"
```

On the command line, execute:

```terminal
flask run
```

And output similar to the following will be provided:

```terminal
Running on http://127.0.0.1:5000/ (Press CTRL+C to quit)
```

Browsing to `http://127.0.0.1:5000/` will result in a page with the text
*Hello!*.

#### Add a Chart

CanvasXpress for Python can be used to define a chart with various attributes and then generate the necessary HTML and
Javascript for proper display in the browser.

Add a `templates` directory to the same location as the `app.py` file, and inside add a file
called `canvasxpress_example.html`. Inside the file add:

```html

<html>
<head>
    <meta charset="UTF-8">
    <title>Flask CanvasXpress Example</title>

    <!-- 2. Include the CanvasXpress library -->
    <link
            href='https://www.canvasxpress.org/dist/canvasXpress.css'
            rel='stylesheet'
            type='text/css'
    />
    <script
            src='https://www.canvasxpress.org/dist/canvasXpress.min.js'
            type='text/javascript'>
    </script>

    <!-- 3. Include script to initialize object -->
    <script type="text/javascript">
        onReady(function () {
            {
                {
                    canvas_source | safe
                }
            }
        })
    </script>

</head>
<body>

<!-- 1. DOM element where the visualization will be displayed -->
{{canvas_element|safe}}

</body>
</html>
```

The HTML file, which uses [Jinja syntax](https://palletsprojects.com/p/jinja/) achieves three things:

1. Provides a location for a `<div>` element that marks where the chart will be placed.
1. References the CanvasXpress CSS and JS files needed to illustrate and operate the charts.
1. Provides a location for the Javascript that will replace the chart `<div>` with a working element on page load.

Going back to our Flask app, we can add a basic chart definition with some data to our example function:

```python
from flask import Flask, render_template
from canvasxpress.canvas import CanvasXpress

app = Flask(__name__)


@app.route('/')
def canvasxpress_example():
    # Define a CX bar chart with some basic data
    chart: CanvasXpress = CanvasXpress(
        data={
            "y": {
                "vars": ["Gene1"],
                "smps": ["Smp1", "Smp2", "Smp3"],
                "data": [[10, 35, 88]]
            }
        },
        config={
            "graphType": "Bar"
        }
    )

    # Get the HTML parts for use in our Web page:
    html_parts: dict = chart.render_to_html_parts()

    # Return a Web page based on canvasxpress_example.html and our HTML parts
    return render_template(
        "canvasxpress_example.html",
        canvas_element=html_parts["cx_canvas"],
        canvas_source=html_parts["cx_js"]
    )
```

#### Run the App and View the Page

Rerun the flask app on the command line and browse to the indicated IP and URL. A page similar to the following will be
displayed:

<img src="https://raw.githubusercontent.com/docinfosci/canvasxpress-python/main/readme/examples/flask_bar_chart_basic.png" align="center" width="600"></a>

Congratulations!  You have created a Flask CanvasXpress app!
</details>
'''

core_pkgs = [
    "deepdiff",
    "gitpython>=3",
    "pytz>=2020.1",
    "pandas>=1.1.5",
    "requests>=2",
    "Deprecated>=1.2.12"
]
dash_pkgs = [
    "dash"
]
jupyter_pkgs = [
    "jupyter",
    "IPython>=7",
    "beautifulsoup4>=4",
    "htmlmin>=0.1.12",
    "requests"
]
streamlit_pkgs = [
    "streamlit"
]
shiny_pkgs = [
    "shiny",
    "htmltools",
    "ipywidgets",
    "shinywidgets"
]
rstudio_pkgs = [
    "rpy2"
]

setup(
    name='canvasxpress',
    version='2026.2.8.001918',
    packages=find_packages(exclude=["tests*", "plotly", "streamlit", "tutorials",]),
    package_data={'': ['*.json', '*.yaml', '*.yml', '*.js', '*.css', '*.html', '*.sql', '*.txt', '*.zip']},
    include_package_data=True,
    package_dir={'': '.'},
    install_requires=core_pkgs,
    extras_require={
        "core": core_pkgs, 
        "dash": core_pkgs + dash_pkgs,
        "streamlit": core_pkgs + streamlit_pkgs,
        "jupyter": core_pkgs + jupyter_pkgs,
        "shiny": core_pkgs + shiny_pkgs,
        "rstudio": core_pkgs + shiny_pkgs + jupyter_pkgs + rstudio_pkgs,
        "all": core_pkgs + dash_pkgs + jupyter_pkgs,
    },
    url='https://github.com/docinfosci/canvasxpress-python.git',
    project_urls={
        'Documentation': 'https://canvasxpress-python.readthedocs.io',
    },
    license='Copyright 2020 to 2026 CanvasXpress all rights reserved',
    author=(
        ' Dr. Isaac Neuhaus for CanvasXpress JS, PHP, and R;'
        ' Dr. Constance M. Brett for R;'
        ' Dr. Todd C. Brett for Python and Dash; and'
        ' Dr. Jennifer Walker for Python.'
    ),
    author_email='todd@aggregate-genius.com',
    description='CanvasXpress for Python',
    long_description=long_description,
    long_description_content_type='text/markdown; charset=UTF-8; variant=GFM',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Environment :: Web Environment',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3',
    ]
)



import os
import tempfile
import uuid
import webbrowser
from copy import deepcopy
from time import sleep
from typing import Any, Union, List

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.base import CXRenderable

_cx_fx_template = """
<script type="text/javascript">
    @code@
</script>
"""

_cx_html_template = """
<html>
    <head>
        <meta charset="UTF-8">
        <title>CanvasXpress</title>

        <!-- 1. Include the CanvasXpress library -->
        @canvasxpress_license@
        <link 
                href='@css_url@' 
                rel='stylesheet' 
                type='text/css'
                referrerpolicy='origin-when-cross-origin'
        />
        <script 
                src='@js_url@' 
                type='text/javascript'>
                referrerpolicy='origin-when-cross-origin'
        </script>
    </head>
    <body>
        <!-- 3. DOM element where the visualization will be displayed -->
        @canvases@
        <!-- 2. Include script to initialize object -->
        @js_functions@
     </body>
</html>
"""


class CXBrowserPopup(CXRenderable):
    """
    CXBrowserPopup is a `CXRenderable` that renders `CanvasXpress` objects into
    a Web page that is displayed in a pop-up browser window.
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXBrowserPopup` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    def render(self, **kwargs: Any):
        """
        Renders the associated CanvasXpress object appropriate for display in
        a pop-up browser window.  Charts cannot have the same name,
        so render_to will be updated with a uuid for each conflicting chart.
        :param kwargs: `Any`
            Supports `columns` for any positive `int` of `1` or greater, with a
            default value of `1`.  Values less that `1` are ignored.  `columns`
            indicates how many charts should be rendered horizontally in the
            browser if more than one chart is being tracked.
        """
        render_targets = list()

        if self.canvas is None:
            pass

        elif isinstance(self.canvas, CanvasXpress):
            render_targets.append(deepcopy(self.canvas))

        else:
            for chart in self.canvas:
                render_targets.append(deepcopy(chart))

        used_render_targets = list()
        for target in render_targets:
            original_render_target = target.render_to
            if original_render_target in used_render_targets:
                target.render_to = (
                    original_render_target + "_" + str(uuid.uuid4()).replace("-", "_")
                )

            used_render_targets.append(target.render_to)

        render_targets.reverse()

        html_parts = [target.render_to_html_parts() for target in render_targets]

        canvases = [part["cx_canvas"] for part in html_parts]

        functions = [part["cx_js"] for part in html_parts]

        cx_license = ""
        for part in html_parts:
            if part.get("cx_license"):
                cx_license = part["cx_license"]
                break

        columns_arg = int(kwargs.get("columns", 1))
        columns = columns_arg if columns_arg > 0 else 1
        if len(canvases) < columns:
            columns = len(canvases)

        page_width = 0
        page_height = 0
        chart_count = len(canvases)
        canvas_table = '<table style="width:100%">'

        while chart_count > 0:
            candidate_width = 0
            candidate_height = 0

            canvas_table += "<tr>"
            for c in range(columns):
                canvas_table += "<td>"
                if chart_count > 0:
                    canvas_table += canvases[chart_count - 1]

                    candidate_width += render_targets[chart_count - 1].width
                    if render_targets[chart_count - 1].height > candidate_height:
                        candidate_height = render_targets[chart_count - 1].height

                canvas_table += "</td>"
                chart_count = chart_count - 1

            canvas_table += "</tr>"

            if candidate_width > page_width:
                page_width = candidate_width
            page_height += candidate_height

        js_functions = "\n".join(
            [_cx_fx_template.replace("@code@", fx) for fx in functions]
        )

        css_url = CanvasXpress.css_library_url()
        js_url = CanvasXpress.js_library_url()

        html = (
            _cx_html_template.replace("@canvases@", canvas_table)
            .replace("@canvasxpress_license@", cx_license)
            .replace("@js_functions@", js_functions)
            .replace("@css_url@", css_url)
            .replace("@js_url@", js_url)
        )

        tempdir = tempfile.TemporaryDirectory()

        temp_filename = os.path.join(tempdir.name, f"{str(uuid.uuid4())}.html")
        with open(temp_filename, "w") as temp_file:
            temp_file.write(html)

        webbrowser.open("file://" + temp_filename, new=1)

        sleep(2)
        tempdir.cleanup()



from urllib.parse import urlparse, urlunparse

from canvasxpress.data.base import CXData


class CXUrlData(CXData):
    """
    CXUrlData provides the ability to accept a URL and pass it properly to the
    CanvasXpress object.
    """

    __url: str = ""
    """
    The tracked URL.
    """

    @property
    def url(self) -> str:
        """
        Provides the URL formatted as a string.
        :returns: `str`
            A string representation of the URL, similer to the data that will
            be provided to the CanvasXpress object.
        """
        return self.__url

    @url.setter
    def url(self, url: str) -> None:
        """
        Tracks the specified URL. `ValueError` will be raised if the URL is
        malformed.
        :param url: `str`
            A string form of the URL, which must be in a stanard format for
            files, http resources, ftp resources, etc.
        """
        if self.validate_url(url):
            self.__url = url

    @property
    def data(self) -> dict:
        """
        A property accessor for the data managed by the object.  Regardless of
        the input data the returned data structure will be a dict-type for use
        with CanvasXpress.
        :returns: `dict`
            A dictionary representing a data map suitable for use with a chart.
        """
        components = urlparse(self.url)
        return components._asdict()

    @data.setter
    def data(self, data: dict) -> None:
        """
        Sets the URL using a data dict with the following attributes:
        `scheme`, `netloc`, `path`, `query`, `fragment`, `username`,
        `password`, `hostname`, `port`, `params`. `ValueError` will be
        raised if the URL is malformed.
        :param data: `dict`
            The URL as broken into a dict with the above attributes set to a
            valid string for the topic or an empty string.
        """
        if not data:
            raise ValueError("data cannot be None.")

        if not isinstance(data, dict):
            raise TypeError("data must be a type of dict")

        candidate = tuple(data)
        self.url = urlunparse(candidate)

    def get_raw_dict_form(self) -> dict:
        """
        Provides a simple dict perspective of the data with no metadata or other
        contextual transformations performed.  For example, if the data is
        natively in `dict` form then it would be passed-through with no
        modification or enhancement.
        :returns: `dict`
            The `dict` perspective of the data with as little modification or
            interpretation as is reasonable.
        """
        return {"raw": self.url}

    def render_to_dict(self, **kwargs) -> dict:
        """
        Converts the object into a dict representation.
        :returns: `dict`
            A dictionary representation of the object, such as what might be
            needed for a JSON export.
        """
        return self.get_raw_dict_form()

    @classmethod
    def validate_url(cls, url: str, detail_errors: bool = True) -> bool:
        """
        Validates that the URL conforms to a recognized standard.  URLs must
        begin with a valid scheme such as `file://` or `http://`.
        :param url: `str`
            The string form of the URL to be validated.
        :param detail_errors: `bool`
            True (default) if an exception should be raised with information
            detailing issues or errors, or `False` if the method should
            return `False` upon encountering issues.
        :returns: `bool`
            `True` if the URL is valid, or `False` if issues are found in the
             URL.  If `detail_errors` is `True` then a ValueError will instead
             be raised with information pertaining to the rejection.
        """
        try:
            components = urlparse(url)
            if not components.scheme:
                raise ValueError(
                    "URLs must specify a scheme (e.g., file://, http://, etc.)."
                )

        except ValueError as ve:
            if detail_errors:
                raise ve
            else:
                return False

        return True

    def __init__(self, data: str) -> None:
        """
        Initializes the CXUrlData object with a valid URL.  URLs must begin with
        a valid scheme such as `file://` or `http://`.  `ValueError` will be
        raised if the URL is malformed.
        :param data: `str`
            A string form of the URL, which must be in a stanard format for
            files, http resources, ftp resources, etc.
        """
        super().__init__(data)
        self.url = data



from abc import ABC, abstractmethod


class CXHtmlConvertable(ABC):
    """
    CXHtmlConvertable represents an object that can be converted into HTML.
    """

    @abstractmethod
    def render_to_html_parts(self) -> dict:
        """
        Converts the object into HTML5 compliant script.
        :returns: `dict`
            A `dict` representation of the object with each component of the
            necessary HTML indicated by a key.  For example, there might be
            a `div` element and a `script` import.
        """
        pass


class CXDictConvertable(ABC):
    """
    CXDictConvertable represents an object that can be converted into a dict.
    """

    @abstractmethod
    def render_to_dict(self, **kwargs) -> dict:
        """
        Converts the object into a dict representation.
        :param kwargs:
            Keyword arguments that can be supplied to facilitate rendering
            decisions.
        :returns: `dict`
            A dictionary representation of the object, such as what might be
            needed for a JSON export.
        """
        pass


class CXListConvertable(ABC):
    """
    CXListConvertable represents an object that can be converted into a list.
    """

    @abstractmethod
    def render_to_list(self, **kwargs) -> list:
        """
        Converts the object into a list representation.
        :param kwargs:
            Keyword arguments that can be supplied to facilitate rendering
            decisions.
        :returns: `list`
            A list representation of the object, such as what might be
            needed for a JSON export.
        """
        pass


class CXJavascriptConvertable(ABC):
    """
    CXJavascriptConvertable represents an object that can be converted into JS.
    """

    @abstractmethod
    def render_to_js(self) -> str:
        """
        Converts the object into HTML5 complant script.
        :returns: `str`
            A string representation of the object in a form that can be used
            within HTML or Javascript.
        """
        pass



from canvasxpress.canvas import CanvasXpress


def generate_canvasxpress_code_from_json_file(
    cx_json_path: str,
    document_includes: bool = True,
    document_render: bool = True,
    document_jupyter_render=False,
) -> str:
    """
    Generates a string with a CanvasXPress in Python declaration using a
    CanvasXpress reproducible research JSON stored in a file.
    :param cx_json_path: `str`
        A valid path to the reproducible JSON text from which a CanvasXPress
        object is to be built and then converted into example code.
    :param document_includes: `bool`
        Default `True`.  Indicate if include headers should be prefixed.
    :param document_render: `bool`
        Default `True`.  Indicate if rendering should be included in the
        example code.
    :param document_jupyter_render: `bool`
        Default `False`.  Indicate if Jupyter rendering should be performed;
        otherwise, popup rendering will suffixed.
    :returns: `str`
        A string with the code example.
    """
    with open(cx_json_path, "r") as cx_json_file:
        cx_json = cx_json_file.read()

        return generate_canvasxpress_code(
            CanvasXpress.from_reproducible_json(cx_json),
            document_includes,
            document_render,
            document_jupyter_render,
        )


def generate_canvasxpress_code_from_json(
    cx_json: str,
    document_includes: bool = True,
    document_render: bool = True,
    document_jupyter_render=False,
) -> str:
    """
    Generates a string with a CanvasXPress in Python declaration using a
    CanvasXpress reproducible research JSON.
    :param cx_json: `str`
        The reproducible JSON text from which a CanvasXPress object is to be
        built and then converted into example code.
    :param document_includes: `bool`
        Default `True`.  Indicate if include headers should be prefixed.
    :param document_render: `bool`
        Default `True`.  Indicate if rendering should be included in the
        example code.
    :param document_jupyter_render: `bool`
        Default `False`.  Indicate if Jupyter rendering should be performed;
        otherwise, popup rendering will suffixed.
    :returns: `str`
        A string with the code example.
    """
    return generate_canvasxpress_code(
        CanvasXpress.from_reproducible_json(cx_json),
        document_includes,
        document_render,
        document_jupyter_render,
    )


def generate_canvasxpress_code(
    cx: CanvasXpress,
    document_includes: bool = True,
    document_render: bool = True,
    document_jupyter_render=False,
) -> str:
    """
    Generates a string with a CanvasXPress in Python declaration.
    :param cx: `CanvasXpress`
        The `CanvasXpress` object from which to generate the example code.
    :param document_includes: `bool`
        Default `True`.  Indicate if include headers should be prefixed.
    :param document_render: `bool`
        Default `True`.  Indicate if rendering should be included in the
        example code.
    :param document_jupyter_render: `bool`
        Default `False`.  Indicate if Jupyter rendering should be performed;
        otherwise, popup rendering will suffixed.
    :returns: `str`
        A string with the code example.
    """
    example_text = ""

    if document_includes:
        example_text += "from canvasxpress.canvas import CanvasXpress \n"
        example_text += "from canvasxpress.js.collection import CXEvents \n"

        if document_render:
            if document_jupyter_render:
                example_text += (
                    "from canvasxpress.render.jupyter" " import CXNoteBook \n"
                )
            else:
                example_text += (
                    "from canvasxpress.render.popup" " import CXBrowserPopup \n"
                )

    example_text += "\n"
    example_text += "cx = " + repr(cx)

    if document_render:
        if document_jupyter_render:
            example_text += "\n"
            example_text += "display = CXNoteBook(cx) \n"

        else:
            example_text += "\n"
            example_text += "display = CXBrowserPopup(cx) \n"

    example_text += "display.render() \n"

    return example_text
